# run with : python3 -m unittest TestPrior
# or :       python3 -m unittest TestPrior.TestPrior.testUniformPrior

import unittest
import numpy as numpy
import math

from BayesicFitting import Prior, UniformPrior, JeffreysPrior, ExponentialPrior
from BayesicFitting import LaplacePrior, CauchyPrior, GaussPrior
from BayesicFitting import CircularUniformPrior


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2006-2015 Do Kester (JAVA CODE)

class TestPrior( unittest.TestCase ) :
    """
    Test harness for Fitter class.

    @author Do Kester

    """
    def testUniformPrior( self ):

        print( "===== Uniform Prior Tests ==========================\n" )
#        self.assertWarns( UserWarning, UniformPrior )
        prior = UniformPrior( )
        print( prior )
        self.assertTrue( prior._lowDomain == -math.inf )
        self.assertTrue( prior._highDomain == +math.inf )
        self.assertFalse( prior.isBound() )
#        self.assertRaises( AttributeError, prior.unit2Domain, 0.0 )
#        self.assertRaises( AttributeError, prior.domain2Unit, 0.0 )
#        self.assertRaises( AttributeError, prior.result, 1.0 )

        prior.setLimits( [0,5] )
        print( "lowlim %f  highlim %f range %f"%( prior.lowLimit, prior.highLimit, prior._urng ) )
        self.assertTrue( prior._urng == prior.getIntegral() )
        print( prior )

        values = {0.0:0.0, 0.5:2.5, 1.0:5.0 }
        self.stdTestPrior( prior, values=values )

        cp = prior.copy( )
        print( cp )
        self.stdTestPrior( cp, values=values )

        prior = UniformPrior( numpy.asarray( [6.,48.] ) )
        print( prior )
        values = {0.0:6, 0.5:27, 1.0:48 }
        self.stdTestPrior( prior, values=values )

        print( prior.domain2Unit( prior.unit2Domain(0.1) ) )
        print( prior.domain2Unit( prior.unit2Domain(0.3) ) )
        print( prior.domain2Unit( prior.unit2Domain(0.8) ) )
        self.assertAlmostEqual( prior.domain2Unit( prior.unit2Domain( 0.1 ) ), 0.1 )
        self.assertAlmostEqual( prior.domain2Unit( prior.unit2Domain( 0.3 ) ), 0.3 )
        self.assertAlmostEqual( prior.domain2Unit( prior.unit2Domain( 0.8 ) ), 0.8 )
        print( prior.partialDomain2Unit( 5.9 ) )
        print( prior.partialDomain2Unit( 10.3 ) )
        print( prior.partialDomain2Unit( 48.8 ) )
        self.assertTrue( prior.partialDomain2Unit( 5.9 ) == 0.0 )
        self.assertTrue( prior.partialDomain2Unit( 10.3 ) == 1.0 / ( 48 - 6) )
        self.assertTrue( prior.partialDomain2Unit( 48.1 ) == 0.0 )
        self.assertAlmostEqual( prior.partialDomain2Unit( 18.9 ), prior.numPartialDomain2Unit( 18.9 ) )

    def testCircularUniformPrior( self ):

        print( "===== CircularUniform Prior Tests ==========================\n" )
#        self.assertWarns( UserWarning, CircularUniformPrior )
        prior = CircularUniformPrior( )
        print( prior )
        self.assertTrue( prior._lowDomain == -math.inf )
        self.assertTrue( prior._highDomain == +math.inf )
        self.assertFalse( prior.isBound() )

        prior.setLimits( [0,5] )
        print( "lowlim %f  highlim %f range %f"%( prior.lowLimit, prior.highLimit, prior._urng ) )
        print( prior )

        values = {0.0:0.0, 0.5:2.5, 1.0:5.0 }
        self.stdTestPrior( prior, values=values, utest=False )

        cp = prior.copy( )
        print( cp )
        self.stdTestPrior( cp, values=values, utest=False )

        prior = UniformPrior( numpy.asarray( [6.,48.] ) )
        print( prior )
        values = {0.0:6, 0.5:27, 1.0:48 }
        self.stdTestPrior( prior, values=values, utest=False )

        print( prior.domain2Unit( prior.unit2Domain(0.1) ) )
        print( prior.domain2Unit( prior.unit2Domain(0.3) ) )
        print( prior.domain2Unit( prior.unit2Domain(0.8) ) )
        self.assertAlmostEqual( prior.domain2Unit( prior.unit2Domain( 0.1 ) ), 0.1 )
        self.assertAlmostEqual( prior.domain2Unit( prior.unit2Domain( 0.3 ) ), 0.3 )
        self.assertAlmostEqual( prior.domain2Unit( prior.unit2Domain( 0.8 ) ), 0.8 )
        print( prior.partialDomain2Unit( 5.9 ) )
        print( prior.partialDomain2Unit( 10.3 ) )
        print( prior.partialDomain2Unit( 48.8 ) )
        self.assertTrue( prior.partialDomain2Unit( 5.9 ) == 0.0 )
        self.assertTrue( prior.partialDomain2Unit( 10.3 ) == 1.0 / ( 48 - 6) )
        self.assertTrue( prior.partialDomain2Unit( 48.1 ) == 0.0 )
        self.assertAlmostEqual( prior.partialDomain2Unit(8.9 ), prior.numPartialDomain2Unit( 8.9 ) )

    def testJeffreysPrior( self ):
        print( "===== Jeffreys Prior Tests ===========================\n" )
#        self.assertWarns( UserWarning, JeffreysPrior )
        prior = JeffreysPrior( )
        print( prior )

        self.assertTrue( prior._lowDomain == 0 )
        self.assertTrue( prior._highDomain == +math.inf )
        self.assertFalse( prior.isBound() )

        prior = JeffreysPrior( limits=[0.01,100] )
        print( prior )
        print( prior._umin, prior._urng )

        self.assertEqual( prior.lowLimit, 0.01 )
        self.assertEqual( prior.highLimit, 100 )
        self.assertTrue( prior._lowDomain == 0 )
        self.assertTrue( prior._highDomain == +math.inf )
        self.assertEqual( prior._umin, math.log( 0.01 ) )
        self.assertEqual( prior._urng, math.log(100) - math.log(0.01) )
        self.assertTrue( prior.isBound() )

        values = {0.0:0.01, 0.5:1.0, 1.0:100.0}
        self.stdTestPrior( prior, values=values )
        self.assertTrue( prior._umin == math.log( 0.01 ) )
        self.assertTrue( prior._urng == math.log( 10000.0 ) )

        prior.setLimits( numpy.asarray( [1.,6.] ) )
        print( prior )
        print( prior._umin, prior._urng )

        self.assertTrue( prior._umin == math.log( 1.0 ) )
        self.assertTrue( prior._urng == math.log( 6.0 ) )
        self.assertTrue( prior._urng == prior.getIntegral() )

        values = {0.0: 1.0, 0.5: 2.449489742783178, 1.0:6.0}
        self.stdTestPrior( prior, values=values )

        cp = prior.copy( )
        print( "Copy : " + str( cp ) )
        self.stdTestPrior( cp, values=values )

        prior = JeffreysPrior( numpy.asarray( [6.,48.] ) )
        print( prior )
        self.assertTrue( prior._umin == math.log( 6.0 ) )
        self.assertTrue( prior._urng == math.log( 48.0 ) - math.log( 6.0 ) )
        values = {0.0:6.0, 1.0:48.0}
        self.stdTestPrior( prior, values=values )


    def stdTestPrior( self, prior, utest=True, values={} ) :
        print( "----- Standard test ----------------------------------------" )

        for ku in range( 11 ) :
            u = 0.1 * ku
            d = prior.unit2Domain( u )
#            uu = u * prior._urng + prior._umin
#            print( ku, u, uu, math.exp( uu ), d, prior.isOutOfLimits( math.exp( uu ) ) )
            v = prior.domain2Unit( d )
            f = prior.unit2Domain( v )
            print( "Unit %10.7f %10.7f  Domain %10.7f %10.7f"%( u, v, d, f ) )
            self.assertAlmostEqual( d, f )
            if utest :
                self.assertAlmostEqual( v, u )

        # define a number of x values
        xx = [-10,-5, -1, 0, 1, 3, 6, 6.1, 48, math.inf]
        for x in xx :
            pl = prior.partialLog( x )
            nl = prior.numPartialLog( x )
            print( "Prior at %.3f is %.3f partialLogPrior is %.3f ~= %.3f"%
                 ( x, prior.result(x), pl, nl ) )
            if not ( math.isnan( pl + nl ) or prior.isOutOfLimits( x - prior.deltaP ) or
                     prior.isOutOfLimits( x + prior.deltaP ) ) :
                self.assertAlmostEqual( pl, nl, 4 )
            self.assertTrue( prior.result( x ) == prior.partialDomain2Unit( x ) )

        for v in values.keys() :
            print( "Unit %10.7f %10.7f  Domain %10.7f %10.7f"%
                ( v, prior.domain2Unit( values[v] ), values[v], prior.unit2Domain( v ) ) )
            self.assertAlmostEqual( prior.unit2Domain( v ), values[v] )
            if utest :
                self.assertAlmostEqual( prior.domain2Unit( values[v] ), v )

    def domainTest( self, prior ) :
        print( "----- Domain test ----------------------------------------" )

        print( prior.unit2Domain( 1.0 ) )
        print( prior.unit2Domain( 0.99999999999999993 ) )
        print( prior.unit2Domain( 0.9999999999999999 ) )
        print( prior.unit2Domain( 0.99999999999999 ) )
        print( prior.unit2Domain( 0.999999999999 ) )
        print( prior.unit2Domain( 0.9999999999 ) )
        print( prior.unit2Domain( 0.75 ) )
        print( prior.unit2Domain( 0.50 ) )
        print( prior.unit2Domain( 0.25 ) )
        print( prior.unit2Domain( 0.0000000001 ) )
        print( prior.unit2Domain( 0.000000000001 ) )
        print( prior.unit2Domain( 0.00000000000001 ) )
        print( prior.unit2Domain( 0.0000000000000001 ) )
        print( prior.unit2Domain( 0.00000000000000005 ) )
        print( prior.unit2Domain( 0.0 ) )



    def testExponentialPrior( self ):
        print( "===== Exponential Prior Tests ============================\n" )
        prior = ExponentialPrior( )
        print( prior )
        self.assertTrue( prior.scale == 1.0 )
        self.assertTrue( prior.zeroFraction == 0 )
        self.assertTrue( prior._lowDomain == 0 )
        self.assertTrue( prior._highDomain == +math.inf )
        self.assertTrue( prior.isBound() )
        self.assertTrue( prior._uval == 0.0 )
        self.assertTrue( prior._shift == 1.0 )

        self.stdTestPrior( prior )
        self.domainTest( prior )

        maxdom = prior.MAXVAL * prior.scale
        print( prior.unit2Domain( 0.5 ) )
        print( prior.unit2Domain( 0.0 ) )
        self.assertAlmostEqual( prior.unit2Domain( 1 - 1.0 / 1024 ), 6.93147180559945 )
        self.assertAlmostEqual( prior.unit2Domain( 0.5 ), 0.693147180559945 )
        self.assertAlmostEqual( prior.unit2Domain( 0.0 ), 0 )
        self.assertTrue( prior.unit2Domain( 1.0 ) == maxdom )

        prior.scale = 10
        print( prior )
        self.stdTestPrior( prior )

        maxdom = prior.MAXVAL * prior.scale
        self.assertTrue( prior.scale == 10.0 )
        self.assertTrue( prior.zeroFraction == 0 )
        self.assertAlmostEqual( prior.unit2Domain( 1 -1.0 / 1024 ), 69.3147180559945 )
        self.assertAlmostEqual( prior.unit2Domain( 0.5 ), 6.93147180559945 )
        self.assertAlmostEqual( prior.unit2Domain( 0.0 ), 0 )
        self.assertTrue( prior.unit2Domain( 1.0 ) == maxdom )

        cp = prior.copy( )
        print( cp )
        self.assertTrue( cp.scale == 10.0 )
        self.assertTrue( cp.zeroFraction == 0 )
        self.assertAlmostEqual( cp.unit2Domain( 1 -1.0 / 1024 ), 69.3147180559945 )
        self.assertAlmostEqual( cp.unit2Domain( 0.5 ), 6.93147180559945 )
        self.assertAlmostEqual( cp.unit2Domain( 0.0 ), 0 )
        self.assertTrue( cp.unit2Domain( 1.0 ) == maxdom )

        self.stdTestPrior( cp )

        self.assertRaises( ValueError, prior.__setattr__, "zeroFraction", 1.3 )
        self.assertRaises( ValueError, prior.__setattr__, "zeroFraction", -0.3 )


        prior.zeroFraction = 0.5
        print( prior )
        self.assertTrue( prior.scale == 10.0 )
        self.assertTrue( prior.zeroFraction == 0.5 )
        self.assertTrue( prior._uval == 0.0 )
        self.assertTrue( prior._shift == 0.5 )

        self.stdTestPrior( prior, utest=False )

        maxdom = prior.MAXVAL * prior.scale
        self.assertAlmostEqual( prior.unit2Domain( 1 - 1.0 / 2048 ), 69.3147180559945 )
        self.assertAlmostEqual( prior.unit2Domain( 0.75 ), 6.93147180559945 )
        self.assertAlmostEqual( prior.unit2Domain( 0.5 ), 0 )
        self.assertAlmostEqual( prior.unit2Domain( 0.25 ), 0 )
        self.assertAlmostEqual( prior.unit2Domain( 0.0 ), 0 )
        self.assertTrue( prior.unit2Domain( 1.0 ) == maxdom )
        print( prior.domain2Unit( 0 ) )
        uv = prior.domain2Unit( 0 )
        self.assertTrue( prior.domain2Unit(prior.unit2Domain(uv) ) != uv )
        print( "partial 0.0 : ", prior.partialDomain2Unit( 0.0 ) )
        print( "numpart 0.0 : ", prior.numPartialDomain2Unit( 0.0 ) )
        print( "partial 0.1 : ", prior.partialDomain2Unit( 0.1 ) )
        print( "partial  1. : ", prior.partialDomain2Unit( 1.0 ) )
        print( "partial 10. : ", prior.partialDomain2Unit( 10. ) )
        self.assertAlmostEqual( prior.partialDomain2Unit(0.0 ), 0.05 )
        print( prior.partialDomain2Unit( 9 ) )
        print( prior.numPartialDomain2Unit( 9 ) )
        self.assertAlmostEqual( prior.partialDomain2Unit(9 ), prior.numPartialDomain2Unit( 9 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(1 ), prior.numPartialDomain2Unit( 1 ), 4 )

        prior.scale = 100.0
        self.assertTrue( prior.scale == 100.0 )

        cp = ExponentialPrior( prior=prior, scale=10 )
        print( cp )
        maxdom = cp.MAXVAL * cp.scale
        self.assertTrue( cp.zeroFraction == 0.5 )
        self.assertTrue( cp.scale == 10 )
        print( cp.domain2Unit( 0 ) )
        self.assertAlmostEqual( cp.unit2Domain( 1 - 1.0 / 2048 ), 69.3147180559945 )
        self.assertAlmostEqual( cp.unit2Domain( 0.75 ), 6.93147180559945 )
        self.assertAlmostEqual( cp.unit2Domain( 0.4 ), 0 )
        self.assertTrue( cp.unit2Domain( 1.0 ) == maxdom )

        prior = ExponentialPrior( 2 )
        print( prior )
        maxdom = prior.MAXVAL * prior.scale
        self.assertTrue( prior.scale == 2 )
        self.assertAlmostEqual( prior.unit2Domain(prior.domain2Unit(0.0) ), 0.0 )
        self.assertAlmostEqual( prior.unit2Domain(prior.domain2Unit(3) ), 3 )
        self.assertAlmostEqual( prior.unit2Domain(prior.domain2Unit(50) ), 50.0, 4 )
        self.assertAlmostEqual( prior.unit2Domain(prior.domain2Unit(100) ), maxdom )

        prior.zeroFraction = 0.3
        print( prior )
        self.assertTrue( prior.scale == 2.0 )
        self.assertTrue( prior.zeroFraction == 0.3 )
        self.assertTrue( prior._uval == 0.0 )
        self.assertTrue( prior._shift == 0.7 )

        self.stdTestPrior( prior, utest=False )
        print( prior._uval )
        print( prior.domain2Unit( 0 ) )
        print( prior._uval )
        uv = prior.domain2Unit( 0 )
        self.assertTrue( uv == prior.domain2Unit( 0) )
        self.assertTrue( prior.domain2Unit(prior.unit2Domain(0.0) ) <= 0.3 )
        self.assertTrue( prior.domain2Unit(prior.unit2Domain(0.1) ) <= 0.3 )
        self.assertTrue( prior.domain2Unit(prior.unit2Domain(0.2) ) <= 0.3 )
        self.assertTrue( prior.domain2Unit(prior.unit2Domain(0.3) ) <= 0.3 )
        self.assertTrue( prior.domain2Unit(prior.unit2Domain(0.7) ) == 0.7 )
        self.assertTrue( prior.domain2Unit(prior.unit2Domain(0.8) ) == 0.8 )
        self.assertTrue( prior.domain2Unit(prior.unit2Domain(0.9) ) == 0.9 )
        self.assertTrue( prior.domain2Unit(prior.unit2Domain(1.0) ) == 1.0 )
        self.assertAlmostEqual( prior.unit2Domain(prior.domain2Unit(0.0) ), 0.0 )

    def testLaplacePrior( self ):
        print( "===== Laplace Prior Tests ==============================\n" )

        prior = LaplacePrior( )
        print( prior )
        self.assertTrue( prior.scale == 1 )
        print( prior.unit2Domain( 0.0 ) )
        print( prior.unit2Domain( 0.25 ) )
        print( prior.unit2Domain( 0.50 ) )
        print( prior.unit2Domain( 0.75 ) )
        print( prior.unit2Domain( 1.0 ) )

        self.stdTestPrior( prior )
        self.domainTest( prior )

        maxdom = prior.center + prior.MAXVAL * prior.scale
        self.assertTrue( prior.unit2Domain(1.0 ) == maxdom )
        self.assertAlmostEqual( prior.unit2Domain(0.75 ), 0.693147180559945 )
        self.assertAlmostEqual( prior.unit2Domain(0.5 ), 0 )
        self.assertAlmostEqual( prior.unit2Domain(0.25 ), -0.693147180559945 )
        self.assertAlmostEqual( prior.unit2Domain(1.0 / 2048 ), -6.93147180559945 )
        self.assertTrue( prior.unit2Domain(0.0 ) == -maxdom )

        prior.scale = 10
        print( prior )
        maxdom = prior.center + prior.MAXVAL * prior.scale
        self.assertTrue( prior.scale == 10 )
        self.assertAlmostEqual( prior.unit2Domain(1.0 / 2048 ), -69.3147180559945 )
        self.assertAlmostEqual( prior.unit2Domain(0.25 ), -6.93147180559945 )
        self.assertAlmostEqual( prior.unit2Domain(0.5 ), 0 )
        self.assertTrue( prior.unit2Domain(1.0 ) == maxdom )

        cp = prior.copy( )
        cp.center = 100
        maxdom = cp.center + cp.MAXVAL * cp.scale
        print( cp )
        self.stdTestPrior( cp )
        self.assertTrue( cp.scale == 10.0 )
        self.assertAlmostEqual( cp.unit2Domain(1.0 / 2048 ), 100 - 69.3147180559945 )
        self.assertAlmostEqual( cp.unit2Domain(0.25 ), 100 - 6.93147180559945 )
        self.assertAlmostEqual( cp.unit2Domain(0.5 ), 100 )
        self.assertTrue( cp.unit2Domain(1.0 ) == maxdom )

        prior.scale = 1.0
        print( prior )
        print( "partial 0.0 : ", prior.partialDomain2Unit( 0.0 ) )
        print( "numpart 0.0 : ", prior.numPartialDomain2Unit( 0.0 ) )
        print( "partial 0.1 : ", prior.partialDomain2Unit( 0.1 ) )
        print( "numpart 0.1 : ", prior.numPartialDomain2Unit( 0.1 ) )
        print( "partial  1. : ", prior.partialDomain2Unit( 1.0 ) )
        print( "numpart  1. : ", prior.numPartialDomain2Unit( 1.0 ) )
        print( "partial 10. : ", prior.partialDomain2Unit( 10. ) )
        print( "numpart 10. : ", prior.numPartialDomain2Unit( 10. ) )
        self.assertAlmostEqual( prior.partialDomain2Unit(0.0 ), 0.5 )
        print( prior.partialDomain2Unit( 9 ) )
        print( prior.numPartialDomain2Unit( 9 ) )
        self.assertAlmostEqual( prior.partialDomain2Unit(-9 ), prior.numPartialDomain2Unit( -9 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(-1 ), prior.numPartialDomain2Unit( -1 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(9 ), prior.numPartialDomain2Unit( 9 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(1 ), prior.numPartialDomain2Unit( 1 ), 4 )

    def testLaplacePrior1( self ):
        print( "===== Laplace Prior with limits ==============================\n" )

        prior = LaplacePrior( limits=[-1.5,None], center=1, scale=2 )
        print( prior )
        print( "u2d   ", prior.unit2Domain )
        print( "bu2d  ", prior.baseUnit2Domain )


        self.assertTrue( prior.scale == 2 )
        print( prior.unit2Domain( 0.0 ) )
        print( prior.unit2Domain( 0.25 ) )
        print( prior.unit2Domain( 0.50 ) )
        print( prior.unit2Domain( 0.75 ) )
        print( prior.unit2Domain( 1.0 ) )

        self.stdTestPrior( prior )
        self.domainTest( prior )

    def testLaplacePrior2( self ):
        print( "===== Circular Laplace Prior  ==============================\n" )

        prior = LaplacePrior( circular=math.pi, center=1, scale=2 )
        print( prior )
        self.assertTrue( prior.scale == 2 )
        print( prior.unit2Domain( 0.0 ) )
        print( prior.unit2Domain( 0.25 ) )
        print( prior.unit2Domain( 0.50 ) )
        print( prior.unit2Domain( 0.75 ) )
        print( prior.unit2Domain( 1.0 ) )

        self.stdTestPrior( prior, utest=False )
        self.domainTest( prior )


    def testCauchyPrior( self ):
        print( "===== Cauchy Prior Tests ===========================\n" )

        prior = CauchyPrior( )
        print( prior )
        self.assertTrue( prior.scale == 1 )

        self.domainTest( prior )

        values = {1.0:1.633123935319537e+16, 0.75:1.0, 0.5:0.0, 0.25:-1.0, 0.0:-1.633123935319537e+16}
        self.stdTestPrior( prior, values=values )

        prior.scale = 10
        print( prior )
        self.assertTrue( prior.scale == 10 )
        self.assertAlmostEqual( prior.unit2Domain( 0.25 ), -9.999999999999998 )
        self.assertAlmostEqual( prior.unit2Domain( 0.5 ), 0 )
        self.assertTrue( prior.unit2Domain( 1.0 ) == 1.633123935319537e+17 )

        cp = prior.copy( )
        print( cp )
        self.assertTrue( cp.scale == 10.0 )
        self.assertAlmostEqual( cp.unit2Domain( 0.25 ), -9.999999999999 )
        self.assertAlmostEqual( cp.unit2Domain( 0.5 ), 0 )
        self.assertTrue( cp.unit2Domain( 1.0 ) == 1.633123935319537e+17 )
        print( prior.domain2Unit( 0 ) )
        uval = [0.1*k for k in range(11)]
        for uv in uval :
            print( uv,  prior.unit2Domain( uv ), prior.domain2Unit( prior.unit2Domain(uv) ) )
            self.assertAlmostEqual( prior.domain2Unit(prior.unit2Domain(uv) ), uv, 10 )

        prior.scale = 1.0
        print( prior )
        print( "partial 0.0 : ", prior.partialDomain2Unit( 0.0 ) )
        print( "numpart 0.0 : ", prior.numPartialDomain2Unit( 0.0 ) )
        print( "partial 0.1 : ", prior.partialDomain2Unit( 0.1 ) )
        print( "numpart 0.1 : ", prior.numPartialDomain2Unit( 0.1 ) )
        print( "partial  1. : ", prior.partialDomain2Unit( 1.0 ) )
        print( "numpart  1. : ", prior.numPartialDomain2Unit( 1.0 ) )
        print( "partial 10. : ", prior.partialDomain2Unit( 10. ) )
        print( "numpart 10. : ", prior.numPartialDomain2Unit( 10. ) )
        self.assertAlmostEqual( prior.partialDomain2Unit( 0.0 ), 0.3183098861837907 )
        print( prior.partialDomain2Unit( 9 ) )
        print( prior.numPartialDomain2Unit( 9 ) )
        self.assertAlmostEqual( prior.partialDomain2Unit(-9 ), prior.numPartialDomain2Unit( -9 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(-1 ), prior.numPartialDomain2Unit( -1 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(9 ), prior.numPartialDomain2Unit( 9 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(1 ), prior.numPartialDomain2Unit( 1 ), 4 )

    def testCauchyPrior1( self ):
        print( "===== Cauchy Prior with limits ==============================\n" )

        prior = CauchyPrior( limits=[-1.5,None], center=1, scale=2 )
        print( prior )
        print( "u2d   ", prior.unit2Domain )
        print( "bu2d  ", prior.baseUnit2Domain )


        self.assertTrue( prior.scale == 2 )
        print( prior.unit2Domain( 0.0 ) )
        print( prior.unit2Domain( 0.25 ) )
        print( prior.unit2Domain( 0.50 ) )
        print( prior.unit2Domain( 0.75 ) )
        print( prior.unit2Domain( 1.0 ) )

        self.stdTestPrior( prior )
        self.domainTest( prior )

    def testCauchyPrior2( self ):
        print( "===== Circular Cauchy Prior  ==============================\n" )

        prior = CauchyPrior( circular=math.pi, center=1, scale=2 )
        print( prior )
        self.assertTrue( prior.scale == 2 )
        print( prior.unit2Domain( 0.0 ) )
        print( prior.unit2Domain( 0.25 ) )
        print( prior.unit2Domain( 0.50 ) )
        print( prior.unit2Domain( 0.75 ) )
        print( prior.unit2Domain( 1.0 ) )

        self.stdTestPrior( prior, utest=False )
        self.domainTest( prior )

    def testGaussPrior( self ):
        print( "===== Gauss Prior Tests ===========================\n" )

        prior = GaussPrior( )
        print( prior )
        self.assertTrue( prior.scale == 1 )

        self.stdTestPrior( prior )
        self.domainTest( prior )

        maxdom = prior.center + prior.MAXVAL * prior.scale
        mindom = prior.center - prior.MAXVAL * prior.scale
        print( "Domain from %f to %f" % (mindom, maxdom ) )
        print( prior.unit2Domain( 1.0 ), prior.unit2Domain( 0.75 ) ) 
        self.assertTrue( prior.unit2Domain( 1.0 ) == maxdom )
        self.assertAlmostEqual( prior.unit2Domain( 0.75 ), 0.4769362762044699 )

        # value for the new (btter version of GaussPrior
#        self.assertAlmostEqual( prior.unit2Domain( 0.75 ), 0.6744897501960818 )
        self.assertAlmostEqual( prior.unit2Domain( 0.5 ), 0 )
        self.assertAlmostEqual( prior.unit2Domain( 0.25 ), -0.4769362762044699 )
        self.assertTrue( prior.unit2Domain( 0.0 ) == mindom )

        prior.scale = 10
        print( prior )
        maxdom = prior.center + prior.MAXVAL * prior.scale
        self.assertTrue( prior.scale == 10 )
        self.assertAlmostEqual( prior.unit2Domain( 0.25 ), -4.769362762044699 )
        self.assertAlmostEqual( prior.unit2Domain( 0.5 ), 0 )
        self.assertTrue( prior.unit2Domain( 1.0 ) == maxdom )

        cp = prior.copy( )
        print( cp )
        cp.center = 100
        maxdom = cp.center + cp.MAXVAL * cp.scale
        mindom = cp.center - cp.MAXVAL * cp.scale
        self.assertTrue( cp.scale == 10.0 )
        print( cp.unit2Domain( 0.25 ), 100 - 6.744897501960818 )
        print( cp.unit2Domain( 0.5 ), 100 )
        self.assertAlmostEqual( cp.unit2Domain( 0.25 ), 100 - 4.769362762044699 )
        self.assertAlmostEqual( cp.unit2Domain( 0.5 ), 100 )
        self.assertTrue( cp.unit2Domain( 1.0 ) == maxdom )
        self.assertTrue( cp.unit2Domain( 0.0 ) == mindom )

        print( prior.domain2Unit( 0 ) )
        uval = [0.1*k for k in range(11)]
        for uv in uval :
            print( uv,  prior.unit2Domain( uv ), prior.domain2Unit( prior.unit2Domain(uv) ) )
            self.assertAlmostEqual( prior.domain2Unit(prior.unit2Domain(uv) ), uv, 10 )

        prior.scale = 1.0
        print( prior )

        print( "result  0.0 : ", prior.result( 0.0 ) )
        print( "partial 0.0 : ", prior.partialDomain2Unit( 0.0 ) )
        print( "numpart 0.0 : ", prior.numPartialDomain2Unit( 0.0 ) )
        print( "partial 0.1 : ", prior.partialDomain2Unit( 0.1 ) )
        print( "numpart 0.1 : ", prior.numPartialDomain2Unit( 0.1 ) )
        print( "partial  1. : ", prior.partialDomain2Unit( 1.0 ) )
        print( "numpart  1. : ", prior.numPartialDomain2Unit( 1.0 ) )
        print( "partial 10. : ", prior.partialDomain2Unit( 10. ) )
        print( "numpart 10. : ", prior.numPartialDomain2Unit( 10. ) )
        self.assertAlmostEqual( prior.partialDomain2Unit(0.0 ), 0.5641895835477563 )

        # Value for new (better) GaussPrior.
        # self.assertAlmostEqual( prior.partialDomain2Unit(0.0 ), 0.3989422804014327 )
        print( prior.partialDomain2Unit( 9 ) )
        print( prior.numPartialDomain2Unit( 9 ) )
        self.assertAlmostEqual( prior.partialDomain2Unit(-9 ), prior.numPartialDomain2Unit( -9 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(-1 ), prior.numPartialDomain2Unit( -1 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(9 ), prior.numPartialDomain2Unit( 9 ), 4 )
        self.assertAlmostEqual( prior.partialDomain2Unit(1 ), prior.numPartialDomain2Unit( 1 ), 4 )

    def testGaussPrior1( self ):
        print( "===== Gauss Prior with limits ==============================\n" )

        prior = GaussPrior( limits=[-1.5,None], center=1, scale=2 )
        print( prior )
        print( "u2d   ", prior.unit2Domain )
        print( "bu2d  ", prior.baseUnit2Domain )


        self.assertTrue( prior.scale == 2 )
        print( prior.unit2Domain( 0.0 ) )
        print( prior.unit2Domain( 0.25 ) )
        print( prior.unit2Domain( 0.50 ) )
        print( prior.unit2Domain( 0.75 ) )
        print( prior.unit2Domain( 1.0 ) )

        self.stdTestPrior( prior )
        self.domainTest( prior )

    def testGaussPrior2( self ):
        print( "===== Circular Gauss Prior  ==============================\n" )

        prior = GaussPrior( circular=math.pi, center=1, scale=2 )
        print( prior )
        self.assertTrue( prior.scale == 2 )
        print( prior.unit2Domain( 0.0 ) )
        print( prior.unit2Domain( 0.25 ) )
        print( prior.unit2Domain( 0.50 ) )
        print( prior.unit2Domain( 0.75 ) )
        print( prior.unit2Domain( 1.0 ) )

        self.stdTestPrior( prior, utest=False )
        self.domainTest( prior )

    @classmethod
    def suite( cls ):
        return ConfiguredTestCase.suite( PriorTest.__class__ )

if __name__ == '__main__':
    unittest.main( )





# run with : python3 -m unittest TestModels.TestModels.testGaussModel

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt
import warnings

from numpy.testing import assert_array_almost_equal as assertAAE
from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestStellarOrbitModel( unittest.TestCase ):
    """
    Test harness for Stellar Orbit Model

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def test1( self ):
        x  = numpy.linspace( 0, 1400, 1401, dtype=float )
        print( "****** STELLAR ORBIT test 1 ***************" )
        m = StellarOrbitModel( spherical=False )
        self.assertTrue( m.getNumberOfParameters( ) == 7 )
        self.assertTrue( m.npbase == 7 )

        # eccen, semimajor, inclin, ascnodepos, ascnodelong, period, periastron
        p = [0.67, 13, 1500.0, 0.0, 0.0, 0.0, 0.0]
        if self.doplot :
            plt.plot( [0], [0], 'k.' )

            hpi = 1.5
            p = [0.67, 13, 1500.0, 0.0, 0.0, 4.0, 0.0]
            for k in range( 5 ) :
                y = m.result( x, p )
                plt.plot( y[:,0], y[:,1], 'k-' )
                plt.plot( y[0,0], y[0,1], 'k.' )
                p[3] += 0.4


            p = [0.67, 13, 1500.0, 0.0, 0.0, 0.0, 0.0]
            for k in range( 5 ) :
                y = m.result( x, p )
                plt.plot( y[:,0], y[:,1], 'r-' )
                plt.plot( y[0,0], y[0,1], 'r.' )
                p[4] += 0.7


            p = [0.67, 13, 1500.0, 0.0, 1.0, 0.0, 0.0]
            for k in range( 5 ) :
                y = m.result( x, p )
                plt.plot( y[:,0], y[:,1], 'g-' )
                plt.plot( y[0,0], y[0,1], 'g.' )
                p[6] += 0.4

            p = [0.67, 13, 1500.0, 0.0, hpi, 0.0, 0.0]
            for k in range( 5 ) :
                y = m.result( x, p )
                plt.plot( y[:,0], y[:,1], 'b-' )
                plt.plot( y[0,0], y[0,1], 'b.' )
                p[5] += 0.4

            plt.show()


    def test2( self ) :
        x  = numpy.linspace( 0, 1400, 7, dtype=float )
        print( "****** STELLAR ORBIT test 2 ***************" )
        m = StellarOrbitModel( )

        # eccen, semimajor, period, periastron, inclin, ascnodepos, ascnodelong
        p = [0.67, 13, 1200.0, 0.1, 0.2, 0.3, 0.4]

        self.dtest( x, m, p )

        print( "Make copy of model: ", m )
        mc = m.copy()

        self.dtest( x, mc, p )

    def test3( self ) :
        x  = numpy.linspace( 0, 1400, 7, dtype=float )
        print( "****** STELLAR ORBIT test 3 ***************" )
        m = StellarOrbitModel( spherical=False )

        # eccen, semimajor, period, periastron, inclin, ascnodepos, ascnodelong
        p = [0.67, 13, 1200.0, 0.1, 0.2, 0.3, 0.4]

        self.dtest( x, m, p )

        print( "Make copy of model: ", m )
        mc = m.copy()

        self.dtest( x, mc, p )

    def test4( self ) :

        print( "****** STELLAR ORBIT test 4 ***************" )
        m = StellarOrbitModel( spherical=False )

        p = math.pi
        rho = numpy.array( [0, 1, 1, 1, 1, 3, 2.4] )
        phi = numpy.array( [0, 0, p, p/2, p/4, p/6, 4.3] )
        rp0 = numpy.append( rho, phi ).reshape(2,-1).transpose()

        xy0 = m.toRect( rp0 )

        print( "x0   ", fma( xy0[:,0] ) )
        print( "y0   ", fma( xy0[:,1] ) )

        rp1 = m.toSpher( xy0 )

        print( "r1   ", fma( rp1[:,0] ) )
        print( "t1   ", fma( rp1[:,1] ) )

        xy1 = m.toRect( rp1 )

        print( "x1   ", fma( xy1[:,0] ) )
        print( "y1   ", fma( xy1[:,1] ) )

        rp2 = m.toSpher( xy1 )

        print( "r2   ", fma( rp2[:,0] ) )
        print( "t2   ", fma( rp2[:,1] ) )

        xy2 = m.toRect( rp2 )

        print( "x2   ", fma( xy2[:,0] ) )
        print( "y2   ", fma( xy2[:,1] ) )

        rp3 = m.toSpher( xy2 )

        print( "r3   ", fma( rp3[:,0] ) )
        print( "t3   ", fma( rp3[:,1] ) )


#        assertAAE( xy0, xy1 )
        assertAAE( xy1, xy2 )
        assertAAE( rp1, rp2 )
        assertAAE( rp2, rp3 )


    def dtest( self, x, m, p ) :

        dfdx = m.baseDerivative( x, p )

        xp = x + 0.0001
        xm = x - 0.0001

        yp = m.result( xp, p )
        ym = m.result( xm, p )

        numx = ( yp - ym ) / 0.0002

        print( "dx0   ", fma( dfdx[:,0] ) )
        print( "nm0   ", fma( numx[:,0] ) )
        print( "dx1   ", fma( dfdx[:,1] ) )
        print( "nm1   ", fma( numx[:,1] ) )
        assertAAE( dfdx, numx )

        part0, part1 = m.basePartial( x, p )
        for k in range( 7 ) :
            pp = p[:]
            pp[k] += 0.0001
            pm = p[:]
            pm[k] -= 0.0001
            yp = m.result( x, pp )
            ym = m.result( x, pm )
            print( "parameter ", k, "  ", m.parNames[k] )

            num0 = ( yp[:,0] - ym[:,0] ) / 0.0002
            num1 = ( yp[:,1] - ym[:,1] ) / 0.0002
            print( "dp0   ", fma( part0[:,k] ) )
            print( "nm0   ", fma( num0 ) )
            print( "dp1   ", fma( part1[:,k] ) )
            print( "nm1   ", fma( num1 ) )
            assertAAE( part0[:,k], num0 )
            assertAAE( part1[:,k], num1 )


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestStellarOrbitModel.__class__ )

if __name__ == '__main__':
    unittest.main( )





from setuptools import setup, find_packages

__version__ = '3.1.0'

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='BayesicFitting',
    version=__version__,
    author='Do Kester',
    author_email='dokester@home.nl',
    install_requires=[
        'numpy >= 1.9',
        'matplotlib >= 2.0',
        'scipy >= 1.0',
        'astropy >= 2.0',
# A modified version of bspline is added to BayesicFitting
#        'bspline>=0.1.1',         # from John Foster and Juha Jeronen at github
        'future'
    ],
    license='LICENSE.txt',
    description='A Python Toolbox for Bayesian fitting.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://www.bayesicfitting.nl",
    keywords = ['Bayesian', 'modeling', 'evidence', 'statistics', 'analysis',
                'regression', 'inference', 'optimization','nested sampling', 'fitting'],
    packages=['BayesicFitting', 'BayesicFitting/source', 'BayesicFitting/source/kernels',
#              'BayesicFitting/alpha',
              'BayesicFitting/test', 'BayesicFitting/examples'],
    classifiers = [
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Topic :: Scientific/Engineering"
    ],
    include_package_data=True,
    zip_safe=True
)



import numpy as numpy
from astropy import units
import math
from . import Tools

from .OrderProblem import OrderProblem

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Alpha"


#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2022 Do Kester


class SalesmanProblem( OrderProblem ):
    """
    Traveling Salesman Problem.

    The parameters give the order in which the nodes are visited.

    The result is a list of distances.

        [dist( x[p[k-1]], x[p[k]] ) for k in range( len( p ) )]

    The number of parameters is equal to the length of the xdata array
    The parameters are initialized at [k for k in range( npars )]

    Examples
    --------
    >>>  tsm = SalesmanProblem( 100 )
    >>> print( tsm )
    TravelingSalesman in 2 dimensions with 100 nodes.
    >>> print( tsm.npars )
    100


    """

    DISNAMES = ["User Defined", "Manhattan", "Euclidic", "Spherical"]


    def __init__( self, xdata=None, weights=None, distance="euclid", scale=None, copy=None ):
        """
        Traveling Salesman problem.


        Parameters
        ----------
        xdata : array_like of shape [np,ndim]
            the nodes to be visited
        weights : array_like
            weights on the arrival nodes
        distance : str or callable
            to calculate the distance between point1 and point2
            "manh"   : Manhattan distance (1 norm) (2 or more dimensions)
            "euclid" : Euclidic (2 norm) (2 or more dimensions) 
            "spher"  : spherical, distance over sphere (2 dimensions only) 
            callable of the form callable( xdata, pars )
        scale : None or float
            scale all distances by this number.
            None : take minimum distance as scale
        copy : SalesmanProblem
            to be copied

        """
        super( ).__init__( xdata=xdata, weights=weights, copy=copy )

        if copy is not None :
            self.distance = copy.distance
            self.iddis = copy.iddis
            self.scale = copy.scale
            return

        if callable( distance ) :
            self.distance = distance
            self.iddis = 0

        elif isinstance( distance, str ) :
            dis = distance.lower()[0]

            self.iddis = 2
            if dis == "e" :
                self.distance = self.euclidic
            elif dis == "m" :
                self.distance = self.manhattan
                self.iddis = 1
            elif dis == "s" :
                self.distance = self.spherical
                self.iddis = 3
            else :
                warnings.warn( "Unknown distance ", distance, " Using euclidic in stead" )
                self.distance = self.euclidic
        else :
            warnings.warn( "Unknown distance ", distance, " Using euclidic in stead" )
            self.distance = self.euclidic

        mindis = self.minimumDistance()
        if mindis < 1e-10 :
            print( "SalesmanProblem. minimumDistance less than 1.e-10" )
            mindis = 1.0

        self.scale = scale if scale is not None else mindis 


    def copy( self ):
        """ Copy method.  """
        return SalesmanProblem( xdata=self.xdata, weights=self.weights, copy=self )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        
        """
        return True

    def result( self, params ):
        """
        Calculates the distance between the nodes (xdata) in the order
        given by the parameters (params), multiplied by the weight at the 
        starting node (if present), divided by the scale

        Each result is 
            res[k] = dis[k] * weight[params[k]] / scale

        Parameters
        ----------
        params : array_like
            values for the parameters.

        Returns
        -------
        An array of distances

        """
        res = self.distance( self.xdata, params ) / self.scale

        if self.weights is None :
            return res 
        else :
            return res * self.weights[params]

    def manhattan( self, xdata, pars, roll=1 ) :
        """
        Use Manhattan distances (1-norm)

        Each distance is 
            dis[k] = SUM_i ( abs( xdata[pars[k],i] - xdata[pars[k+roll],i] ) )

        Parameters
        ----------
        xdata : array-like of shape (ndata,ndim) 
            positional info in several dimensions
        pars : list of indices
            designating the order of the nodess
        roll : int
            number of positions to roll 
        """
        xd = xdata[pars]
        r = numpy.abs( xd - numpy.roll( xd, -roll, axis=0 ) )
        return numpy.sum( r, 1 )

    def euclidic( self, xdata, pars, roll=1 ) :
        """
        Use Euclidic distances (2-norm)

        Each distance is 
            dis[k] = sqrt( SUM_i ( ( xdata[pars[k],i] - xdata[pars[k+roll],i] )^2 ) )

        Parameters
        ----------
        xdata : array-like of shape (ndata,ndim) 
            positional info in several dimensions
        pars : list of indices
            designating the order of the nodes
        roll : int
            number of positions to roll 
        """
        xd = xdata[pars]
        r = numpy.square( xd - numpy.roll( xd, -roll, axis=0 ) )
        return numpy.sqrt( numpy.sum( r, 1 ) )

    def spherical( self, xdata, pars, roll=1 ) :
        """
        Use distances over a 2-d unit sphere.

        Each distance is calculated according to the Haversine formula.
 
        It is assumed that the xdata is in decimal degrees: [longitude, latitude]

        The results are in radian.

        Parameters
        ----------
        xdata : array-like of shape (ndata,2) 
            longitude, latitude info
        pars : list of indices
            designating the order of the nodes
        roll : int
            number of positions to roll 
        """
        if not hasattr( self, "coslat1" ) :

            xd = xdata * math.pi / 180
            xh = xd / 2

            self.coslat1 = numpy.cos( xd[:,1] )

            self.shlon1 = numpy.sin( xh[:,0] )
            self.chlon1 = numpy.cos( xh[:,0] )
            self.shlat1 = numpy.sin( xh[:,1] )
            self.chlat1 = numpy.cos( xh[:,1] )


        cslat1 = self.coslat1[pars]

        shlon1 = self.shlon1[pars]
        chlon1 = self.chlon1[pars]
        shlat1 = self.shlat1[pars]
        chlat1 = self.chlat1[pars]

        cslat2 = numpy.roll( cslat1, -roll )
        shlon2 = numpy.roll( shlon1, -roll )
        chlon2 = numpy.roll( chlon1, -roll )
        shlat2 = numpy.roll( shlat1, -roll )
        chlat2 = numpy.roll( chlat1, -roll )

        a = ( numpy.square( shlat2 * chlat1 - shlat1 * chlat2 ) + 
              cslat1 * cslat2 *
              numpy.square( shlon2 * chlon1 - shlon1 * chlon2 ) )

        # catch round-off error above 1.
        a = numpy.where( a > 1, 1.0, a )

        dis = 2 * numpy.arctan2( numpy.sqrt( a ), numpy.sqrt( 1 - a ) )

        return dis 

    def minimumDistance( self ) :
        """
        Return the smallest distance in the data.

        """        
        ndata = len( self.xdata[:,0] )
        pars = numpy.arange( ndata, dtype=int )
        md = self.distance( self.xdata, pars ).min()
        for roll in range( 2, ndata ) :
            md = self.distance( self.xdata, pars, roll=roll ).min( initial=md )

        return md

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "TravelingSalesman in %d dimensions with %d nodes. %s distance" %
                    ( self.ndim, self.npars, self.DISNAMES[self.iddis] ) )







__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  * 2013 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester


class Kernel( object ):
    """
    A kernel is an even real-valued integrable function.

    It is satisfying the following two requirements:

    1. The integral over [-Inf,+Inf] exists ( < Inf ).
    2. It is an even function: K( x ) = K( -x ).

    A kernel is called bound when it is 0 everywhere
    except when |x| < range < inf.

    All kernels are scaled such that its value at x=0 is 1.0.


    Several kernel functions, K( x ) are defined in this package:

    Name        Definition          Integral    FWHM    range   comment
    Biweight    ( 1-x^2 )^2            16/15    1.08      1.0   aka Tukey
    CosSquare   cos^2( 0.5*PI*x )        1.0    1.00      1.0
    Cosine      cos( 0.5*PI*x )         4/PI    1.33      1.0
    Gauss       exp( -0.5*x*x )   sqrt(2*PI)    1.22      inf
    Huber       min( 1, 1/|x| )          inf    4.00      inf   improper kernel; aka Median
    Lorentz     1 / ( 1 + x*x )           PI    2.00      inf
    Parabola    1 - x*x                  4/3    1.41      1.0
    Sinc        sin(x) / x               1.0    1.21      1.0
    Triangle    1 - |x|                  1.0    1.00      1.0
    Tricube     ( 1 - |x|^3 )^3        81/70    1.18      1.0
    Triweight   ( 1 - x^2 )^3          32/35    0.91      1.0
    Uniform     1.0                      2.0    2.00      1.0   aka Clip

    Tophat 0    1.0                      1.0    1.00      0.5   like Uniform
           1    1 - |x|                  1.0    1.00      1.0   like Triangle
           2    2nd order polynomials    1.0    1.26      1.5
           3    3rd order polynomials    1.0    1.44      2.0
           4    4th order polynomials    1.0    1.60      2.5
           5    5th order polynomials    1.0    1.73      3.0
           6    6th order polynomials    1.0    1.86      3.5

    For all bound Kernels the definition in the table is true for |x| < range;
    elsewhere it is 0.

    Huber is not a proper Kernel as the integral is inf. However it is important
    in robust fitting (RobustShell) to get a madian-like solution for the outliers.

    Attributes
    ----------
    integral : float
        the integral over the valid range
    fwhm : float
        the full width at half maximum
    range : float
        the region [-range..+range] where the kernel is non-zero.

    Author:      Do Kester

    Category:    mathematics/Fitting

    """
    def __init__( self, integral=1.0, fwhm=1.0, range=1.0  ) :
        """
        Constructor

        Parameters
        ----------
        integral : float
            over [-inf, +inf]
        fwhm : float
            full width at half maximum
        range : float
            the region [-range,+range] where the kernel is nonzero
        """

        self.integral = integral
        self.fwhm = fwhm
        self.range = range

    def result( self, x ):
        """
        Return the result for one input value.

        Parameters
        ----------
        x : array-like
            the input value

        """
        pass

    def resultsq( self, xsq ):
        """
        Return the result for one input value.

        Parameters
        ----------
        x : array-like
            the input value
        Parameters: x the square of the input value

        """
        pass

    def partial( self, x ):
        """
        Return the partial derivative wrt input value.

        Parameters
        ----------
        x : array-like
            the input value

        """
        pass

    def isBound( self ):
        """
        Return true when the kernel is bound.
        All non-zero values are between -1 and +1

        """
        pass

    def name( self ):
        """
        Return the name of the kernel.
        """
        return "Kernel of unknown function"


    def __str__( self ):
        """ Return a name.  """
        return self.name()






import numpy as numpy
from astropy import units
import math
import warnings
from . import Tools
from .Formatter import formatter as fmt
from .Formatter import fma

from .Engine import Engine
from .Engine import DummyPlotter

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2022 Do Kester

class GalileanEngine( Engine ):
    """
    Move all parameters in forward steps, with optional mirroring on the edge.

    Move the parameters in a random direction for N iterations; mirror the direction
    on the gradient of the logLikelihood when the parameters enter the zone of logLlow.

    Attributes
    ----------
    nstep : int (5)
        average number of steps to be taken

    Attributes from Engine
    ----------------------
    walkers, errdis, maxtrials, slow, rng, report, unitRange, unitMin, verbose

    Author       Do Kester.

    """

    SIZE = 0.5

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, **kwargs ):
        """
        Default Constructor.

        Parameters
        ----------
        walkers : WalkerList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : GalileanEngine
            to be copied
        kwargs : for Engine
            "slow", "seed", "verbose"

        """
        super( ).__init__( walkers, errdis, copy=copy, **kwargs )

        self.nstep = 5
        self.size = self.SIZE

        self.plotter = DummyPlotter( )

    def copy( self ):
        """ Return copy of this.  """
        engine = GalileanEngine( self.walkers, self.errdis, copy=self )
        engine.nstep = self.nstep

        return engine

    def __str__( self ):
        """ Return the name of this engine.  """
        return str( "GalileanEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False, iteration=0 ):
        """
        Execute the engine by diffusing the parameters.

        Parameters
        ----------
        kw : int
            index in walkerlist, of the walker
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            set walker in place of append

        Returns
        -------
        int : the number of successfull moves

        """

        self.reportCall()

        walker = self.walkers[kw].copy()
        problem = walker.problem
        Lhood = walker.logL
        allpars = walker.allpars
        fitIndex = walker.fitIndex

        npout = 0
        inside = 0
        Ltry = 0
        size = self.size

        self.plotter.start( param=allpars )
        um = UnitMovements( walker, self, size )

        nstep = int( self.nstep * ( 1 + self.rng.rand() ) )
#        maxtrial = self.maxtrials * nstep
        maxtrial = self.maxtrials

        ptry = allpars.copy()
        if self.verbose > 4 :
            print( "LogL  ", fmt( Lhood ), " LowL  ", fmt( lowLhood), nstep, maxtrial, size )
            print( "alpar ", fma( ptry, linelength=200 ) )
            fip = allpars[fitIndex]
            print( "uap   ", fma( self.domain2Unit( problem, fip, fitIndex ), linelength=200) )
            print( "unitr ", fma( self.unitRange, linelength=200 ), self.unitRange[-2] )

        step = 0
        trial = 0

        while True:
            trial += 1
            f = -1

            um.setParameters( problem, allpars )
            if inside == 0 :                            # safely inside lowLhood area
                ptry[fitIndex] = um.stepPars( 1.0 )

            elif inside == 1 :                          # first time outside -> mirror
                f = ( 1.0 if Lhood == Ltry else
                      ( Lhood - lowLhood ) / ( Lhood - Ltry ) )     # lin interpolation to edge

                pedge = ptry.copy()
                pedge[fitIndex] = um.stepPars( f )                # ptry on edge

                dLdp = self.errdis.partialLogL( problem, pedge, fitIndex )
                self.plotter.move( allpars, pedge, col=1, sym=4 )

                um.mirrorOnLowL( dLdp )
                ptry[fitIndex] = um.stepPars( 1 - f )

                self.plotter.move( pedge, ptry, col=2 )
            else:                                       # mirroring failed; do reverse
#                size *= 0.7
                um.reverseVelocity( size )
                ptry[fitIndex] = um.stepPars( 1.0 )

                self.plotter.move( allpars, ptry, col=3, sym=0 )

            ## future extension #########################
            # if self.constrain :
            #     xdata = self.errdis.xdata
            #     ptry = self.constrain( model, ptry, xdata )
            #############################################

            Ltry = self.errdis.logLikelihood( problem, ptry )

            if self.verbose > 4 :
                print( "uval  ", fma( um.upar, linelength=200 ) )
#                print( "Ltry  ", Ltry, "  Lowl  ", lowLhood, self.walkers[-1].logL )

            if Ltry >= lowLhood:
                self.plotter.move( allpars, ptry, col=0, sym=0 )

                allpars = ptry.copy( )
                Lhood = Ltry
                self.reportSuccess( )
                npout = len( allpars )
                inside = 0
                step += 1
                trial = 0

                ## update the walker
                update = len( self.walkers ) if append else kw
                self.setWalker( update, problem, allpars, Lhood, fitIndex=fitIndex )

            else:
                inside += 1
                if inside == 1:
                    self.plotter.move( allpars, ptry, col=5, sym=4 )
                    self.reportReject( )
                else:
                    self.plotter.move( allpars, ptry, col=4, sym=4 )
                    self.reportFailed( )

            if self.verbose > 4 :
                print( "GEng  ", fma(ptry, linelength=200) )
                print( "Ltry  ", fmt(Ltry), inside, step, trial, fmt( size ) )


            if not ( step < nstep and trial < maxtrial ):
                break



        if npout == 0 :
            self.size = max( 0.9 * self.size, 1e-6 )
            if self.verbose > 4 :
                warnings.warn( "GalileanEngine: no steps found" )
        else :
            self.size = min( self.SIZE, self.size * 1.1 )

        self.plotter.stop( param=allpars, name="GalileanEngine" )
        return npout

class UnitMovements( object ):
    """
    Define all parameter movements in unitspace.



    """
    def __init__( self, walker, engine, size ):
        self.problem = walker.problem
        self.fitIndex = walker.fitIndex

        self.np = len( self.fitIndex )
        self.engine = engine
        self.setParameters( self.problem, walker.allpars )

        if self.np > len( self.engine.unitRange ) :
            self.engine.unitRange = numpy.ones( self.np, dtype=float )

        self.upran = self.engine.unitRange[self.fitIndex]
        self.setVelocity( size )

    def uniform( self ) :
        """
        Return random step between [0.5,0.5]
        """
        return self.engine.rng.rand( self.np ) - 0.5

    def setParameters( self, problem, allpars ):
        """
        Set unit values for the applicable parameters
        """
        fip = allpars[self.fitIndex]
        self.upar = self.engine.domain2Unit( problem, fip, kpar=self.fitIndex )

    def stepPars( self, f ):
        """
        Return a new stepped parameter set.
        """
        self.uvel *= self.perf
        uv = self.uvel
        # add small random contribution to velocity
#        uv *= 1 + ( self.uniform() * 0.1 )
        pv = self.upar + uv * f

        # check if outside [0,1]
        pv, uv = numpy.where( pv <= 0, ( -pv, -uv ), ( pv, uv ) )
        self.upar, self.uvel = numpy.where( pv >= 1, ( 2 - pv, -uv ), ( pv, uv ) )

        return self.engine.unit2Domain( self.problem, self.upar, kpar=self.fitIndex )

    def mirrorOnLowL( self, dLdp ):
        """
        Mirror the velocity on the low likelihood border.
        """
        inprod = numpy.sum( dLdp * self.uvel )
        sumsq  = numpy.sum( dLdp * dLdp )

        if sumsq == 0 : return

        self.uvel -= 2 * dLdp * inprod / sumsq

    def setVelocity( self, size ):
        """
        Set a (random) unit velocity.
        """
        self.uvel = self.uniform() * self.upran * size
        self.perf = 1 + self.uniform() * 0.1

    def reverseVelocity( self, size ):
        """
        Reverse the unit velocity when mirroring fails.
        Add small perturbation.
        """
        upv = self.uvel                 # keep original velocity
        self.setVelocity( size )        # get a new one to perturb
        nm = len( self.engine.walkers )
        self.uvel = size * ( self.np * self.uvel - nm * upv ) / ( nm + self.np )






# run with : python3 -m unittest TestEngine3

import unittest
import numpy as numpy
from astropy import units
import math

from BayesicFitting import *
from BayesicFitting import formatter as fmt
from Sphere import Sphere

import matplotlib.pyplot as plt


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestEngine3( unittest.TestCase ):
    """
    Test harness for Engines

    Author       Do Kester

    """
    def __init__( self, name ):
        super( TestEngine3, self ).__init__( name )


    def initEngine( self, order=1, np=101 ):
        m = PolynomialModel( order )

        up1 = UniformPrior( limits=[-1.0,1.0])
        up2 = UniformPrior( limits=[-1.0,1.0] )

        m.priors = [up1, up2]

        xdata = numpy.linspace( -1.0, 1.0, np, dtype=float )
        data = -0.4 * xdata + 0.5

        numpy.random.seed( 345345 )
        numpy.set_printoptions( precision=3, suppress=True )
        return ( m, xdata, data )

    def plotRandomEngine( self ):
        self.testRandomEngine(plot=True )

    def testRandomEngine( self, plot=False ):
        print( "\n   Random Engine Test\n" )
        self.stdenginetest( RandomEngine, nsamp=10, plot=plot )

    def plotGibbsEngine( self ):
        self.testGibbsEngine( plot=True )

    def testGibbsEngine( self, plot=False ):
        print( "\n   Gibbs Engine Test\n" )
        self.stdenginetest( GibbsEngine, nsamp=100, seed=789, plot=plot )
        self.stdenginetest( GibbsEngine, nsamp=10, seed=456, plot=plot )
        self.stdenginetest( GibbsEngine, nsamp=10, seed=123, plot=plot )

    def plotGibbsEngine2( self ):
        self.testGibbsEngine2( plot=True )

    def testGibbsEngine2( self, plot=False ):
        print( "\n   Gibbs Engine Test 2\n" )
        self.stdenginetest2( GibbsEngine, nsamp=10, lowL=-200, plot=plot )
        self.stdenginetest2( GibbsEngine, nsamp=10, lowL=-170, plot=plot )
        self.stdenginetest2( GibbsEngine, nsamp=10, lowL=-165, plot=plot )
        self.stdenginetest2( GibbsEngine, nsamp=10, lowL=-163, plot=plot )

    def plotStepEngine( self ):
        self.testStepEngine( plot=True )

    def testStepEngine( self, plot=False ):
        print( "\n   Step Engine Test\n" )
        self.stdenginetest( StepEngine, nsamp=100, seed=321, plot=plot )
        self.stdenginetest( StepEngine, nsamp=100, seed=654, plot=plot )
        self.stdenginetest( StepEngine, nsamp=100, seed=987, plot=plot )
        self.stdenginetest( StepEngine, nsamp=100, seed=100, plot=plot )

    def plotGalileanEngine( self ):
        self.testGalileanEngine( plot=True )

    def testGalileanEngine( self, plot=False ):
        print( "\n   Galilean Engine Test\n" )
        self.stdenginetest( GalileanEngine, nsamp=100, seed=121, plot=plot )
        self.stdenginetest( GalileanEngine, nsamp=100, seed=132, plot=plot )
        self.stdenginetest( GalileanEngine, nsamp=100, seed=142, plot=plot )

    def plotGalileanEngine2( self ):
        self.testGalileanEngine2( plot=True )

    def testGalileanEngine2( self, plot=False ):
        print( "\n   Galilean Engine Test 2\n" )
        self.stdenginetest2( GalileanEngine, nsamp=10, lowL=-200, plot=plot )
        self.stdenginetest2( GalileanEngine, nsamp=10, lowL=-170, plot=plot )
        self.stdenginetest2( GalileanEngine, nsamp=10, lowL=-165, plot=plot )
        self.stdenginetest2( GalileanEngine, nsamp=10, lowL=-163, plot=plot )

    def testGalileanEngine3( self, plot=False ):
        print( "\n   Galilean Engine Test 3\n" )
#        self.stdenginetest3( GalileanEngine, nsamp=1000, order=2, plot=2 )
#        self.stdenginetest3( GalileanEngine, nsamp=4, order=10, plot=2 )
        self.stdenginetest3( GalileanEngine, nsamp=100, order=100, plot=1 )


    def plotChordEngine( self ):
        self.testChordEngine( plot=True )

    def testChordEngine( self, plot=False ):
        print( "\n   Chord Engine Test\n" )
        self.stdenginetest( ChordEngine, nsamp=100, seed=4213, plot=plot )
        self.stdenginetest( ChordEngine, nsamp=100, seed=4214, plot=plot )
        self.stdenginetest( ChordEngine, nsamp=100, seed=4215, plot=plot )

    def plotChordEngine2( self ):
        self.testChordEngine2( plot=True )

    def testChordEngine2( self, plot=False ):
        print( "\n   Chord Engine Test 2\n" )
        self.stdenginetest2( ChordEngine, nsamp=10, lowL=-200, plot=plot )
        self.stdenginetest2( ChordEngine, nsamp=10, lowL=-170, plot=plot )
        self.stdenginetest2( ChordEngine, nsamp=10, lowL=-165, plot=plot )
        self.stdenginetest2( ChordEngine, nsamp=10, lowL=-163, plot=plot )

    def testChordEngine3( self, plot=False ):
        print( "\n   Chord Engine Test 3\n" )
#        self.stdenginetest3( ChordEngine, nsamp=10, order=2, plot=1 )
#        self.stdenginetest3( ChordEngine, nsamp=10, order=10, plot=1 )
        self.stdenginetest3( ChordEngine, nsamp=100, order=100, plot=1 )

    def testCompare3( self, plot=False ):
        print( "\n   Compare Galilean and Chord Engine Test 3\n" )
        self.stdenginetest3( GalileanEngine, nsamp=100, order=2, plot=1 )
        self.stdenginetest3( ChordEngine, nsamp=100, order=2, plot=1 )
        self.stdenginetest3( GalileanEngine, nsamp=100, order=10, plot=1 )
        self.stdenginetest3( ChordEngine, nsamp=100, order=10, plot=1 )
        self.stdenginetest3( GalileanEngine, nsamp=100, order=100, plot=1 )
        self.stdenginetest3( ChordEngine, nsamp=100, order=100, plot=1 )

    def testEngine( self, myengine, plot=False ) :
        print( "\n   Uniform Engine Test\n" )

        numpy.random.seed( 123456789 )

        np = 10     ## number of parameters
        knots = numpy.arange( np+1, dtype=float )
        mdl = SplinesModel( knots=knots, order=0 )
        mdl.setLimits( lowLimits=-2, highLimits=2 )

        nd = 10
        nx = nd * np
        xdata = ( numpy.arange( nx, dtype=float ) + 0.5 ) / nd
        yd = numpy.random.rand( nx )
        yd -= numpy.mean( yd )
        ydata = numpy.arange( 0, dtype=float )
        for k in nd :
            ydata = numpy.append( ydata, yd )
        problem = ClassicProblem( xdata=xdata, model=mdl, ydata=ydata )
        errdis = GaussErrorDistribution( scale=1.0 )

        allpars = numpy.ones( np, dtype=float )
        Llow = errdis.logLikelihood( problem, allpars )

        allpars = numpy.random.rand( np ) * 2 - 1
        fitIndex = [k for k in range( np )]
        wl = WalkerList( problem, 2, allpars, fitIndex )




    def stdenginetest3( self, myengine, nsamp=100, order=1, plot=0 ) :
        m = Sphere( order )
        m.priors = [ UniformPrior( limits=[-1.0,1.0] ) ]
        xdata = numpy.array( [0.0] )
        data = numpy.array( [0.0] )

        sigma = 1.0
        problem = ClassicProblem( xdata=xdata, model=m, ydata=data )
        errdis = GaussErrorDistribution( scale=sigma )

        allpars = numpy.array( [0.0] * (m.npars) + [sigma] )
        nap = m.npars + 1
        fitIndex = [k for k in range( m.npars )]
        wl = WalkerList( problem, 2, allpars, fitIndex )

        ## determine lowLhood at edge of unit sphere
        allpars[0] = 1.0
        Llow = errdis.logLikelihood( problem, allpars )

        ## starting position
        allpars[0] = 0.9
        wl[0].allpars = allpars.copy()
        wl[0].logL = errdis.logLikelihood( problem, allpars )
        print( nsamp, order, fmt(allpars), Llow, wl[0].logL )

        if plot == 1 :
            plt.figure( order, figsize=[6,6] )

        engine = myengine( wl, errdis )
        engine.nstep = 15
        engine.size = 1.0
        if plot == 2 :
            engine.plotter = Plotter()
        engine.verbose = 5 if nsamp < 10 else 0

        engine.calculateUnitRange()
        engine.unitRange = numpy.array( [1.0] * m.npars )
        engine.unitMin = numpy.array( [0.0] * m.npars )
        engine.unitMean = numpy.array( [0.5] * m.npars )
        engine.maxtrials = 20

        count = numpy.zeros( order, dtype=int )

        rh = math.pow( 0.5, 1/order )
        kp0 = 0
        kp1 = 1
        rcnt = 0
        for k in range( nsamp ) :
            w = wl[0]
            w.allpars = allpars.copy()
            engine.execute( 0, Llow )
            pars = wl[0].allpars[:-1]
            if plot == 1 :
                plt.plot( pars[kp0], pars[kp1], 'r.' )

            count = numpy.where( pars > 0, count+1, count )
            rad = math.sqrt( numpy.sum( numpy.square( pars ) ) )
            if rad > rh :
                rcnt += 1
#            print( rad )
            self.assertTrue( rad <= 1.0000001 )

        engine.printReport()
        print( engine.errdis.ncalls, engine.errdis.nparts )
        print( "Larger  ", count, rcnt, rh )
        print( "Smaller ", nsamp - count, nsamp - rcnt )
        if plot == 1 :
            phi = numpy.linspace( 0, 2 * math.pi, 361, dtype=float )
            xcp = numpy.cos( phi )
            ysp = numpy.sin( phi )

            plt.plot( xcp, ysp, 'k-' )
            plt.plot( rh*xcp, rh*ysp, 'k-' )
            plt.xlabel( "Param %d" % kp0 )
            plt.ylabel( "Param %d" % kp1 )

            plt.xlim( -1.0, 1.0 )
            plt.ylim( -1.0, 1.0 )
            plt.show()


    def stdenginetest( self, myengine, nsamp=4, seed=4213, plot=False ) :
        m, xdata, data = self.initEngine()
        sigma = 2.0
        problem = ClassicProblem( xdata=xdata, model=m, ydata=data )
        errdis = GaussErrorDistribution( )
        errdis.keepFixed( {0: sigma} )
#        Tools.printclass( m )

#        print( m.basePartial( xdata, [0.0,0.0] ) )

        allpars = [0.0,0.0,sigma]
        wl = WalkerList( problem, nsamp, allpars, [0,1]  )
        map = numpy.ndarray( (21,21), dtype=float )
        mmx = -math.inf
        for k0 in range( 21 ) :
            for k1 in range( 21 ) :
                pl = [k0/10-1, k1/10-1, sigma]
                map[k1,k0] = errdis.logLikelihood( problem, pl )
                if map[k1,k0] > mmx :
                    mmx = map[k1,k0]
                    mk1 = k1/10 - 1
                    mk0 = k0/10 - 1

#        print( map )
#        print( mk0, mk1, mmx )

        ay = numpy.linspace( -1, 1, 21 )
        ax = numpy.linspace( -1, 1, 21 )
        v = [-180, -170, -165, -163]

        lowL = -170

        engine = StartEngine( wl, errdis )
        for ks in range( nsamp ) :
            samp = wl[ks]
            while True :
                engine.execute( samp.id, -math.inf )
#                print( fmt( ks ), fmt( wl[ks].logL ), fmt( samp.id ), fmt( samp.logL ) )
                if wl[ks].logL > lowL : break

        engine = myengine( wl, errdis, seed=seed )
        engine.plotter = Plotter()
        engine.nstep = 4
        engine.debug = True
#        engine.verbose = 5

        pevo = wl.getParameterEvolution()
        for k in range( 5 ) :
            engine.plotter.iter = 1
            plt.figure( str( myengine ), figsize=(4,4) )
            plt.contour( ax, ay, map, v )
            plt.plot( [mk0], [mk1], 'k*' )


            klo = engine.rng.randint( 0, nsamp )

            p0 = [w.allpars[0] for w in wl]
            p1 = [w.allpars[1] for w in wl]
            plt.plot( p0, p1, 'm.' )

            p0 = wl[klo].allpars[:2]
            plt.plot( p0[0], p0[1], 'k.' )


            engine.calculateUnitRange()

            engine.execute( wl[klo].id, lowL )
            p1 = wl[klo].allpars[:2]

            print( klo, fmt( p1 ), fmt( wl[klo].logL ) )
            plt.show()


    def stdenginetest2( self, myengine, nsamp=4, lowL=-170, plot=False ) :
        m, xdata, data = self.initEngine()
        sigma = 2.0
        problem = ClassicProblem( xdata=xdata, model=m, ydata=data )
        errdis = GaussErrorDistribution( )
        errdis.keepFixed( {0: sigma} )

        allpars = [0.0,0.0,sigma]
        wl = WalkerList( problem, nsamp, allpars, [0,1]  )

        engine = StartEngine( wl, errdis )
        for ks in range( nsamp ) :
            samp = wl[ks]
            while True :
                engine.execute( samp.id, -math.inf )
#                print( fmt( ks ), fmt( wl[ks].logL ), fmt( samp.id ), fmt( samp.logL ) )
                if wl[ks].logL > lowL : break

        pevo = wl.getParameterEvolution()
#        print( fmt( pevo ) )
        pmin = numpy.min( pevo, 0 )
        pmax = numpy.max( pevo, 0 )
        pran = ( pmax - pmin ) / 2
#        print( fmt( pmin ), fmt( pmax ), fmt( pran ) )
        pmin -= pran
        pmax += pran

        map = numpy.ndarray( (21,21), dtype=float )
        mmx = -math.inf
#        print( fmt(pmin), fmt(pmax) )
#        print( fmt(m.priors[0].lowLimit), fmt(m.priors[0].highLimit) )
#        print( fmt(m.priors[1].lowLimit), fmt(m.priors[1].highLimit) )

        ax = numpy.linspace( max( pmin[0], m.priors[0].lowLimit ),
                min( pmax[0], m.priors[0].highLimit ), 21 )
        ay = numpy.linspace( max( pmin[1], m.priors[1].lowLimit ),
                min( pmax[1], m.priors[1].highLimit ), 21 )

        for k0,p0 in enumerate( ax ) :
            for k1,p1 in enumerate( ay ) :
                pl = [p0, p1, sigma]
                map[k1,k0] = errdis.logLikelihood( problem, pl )
                if map[k1,k0] > mmx :
                    mmx = map[k1,k0]
                    mk1 = p1
                    mk0 = p0

#        print( map )
        print( mk0, mk1, mmx )



        engine = myengine( wl, errdis )
        nstep = 10
        plt.figure( engine.__str__(), figsize=[6,6] )
        v = [-200, -170, -165, -163]
        plt.contour( ax, ay, map, v )
#        plt.plot( [mk0], [mk1], 'k*' )
        plt.xlabel( "Param_0" )
        plt.ylabel( "Param_1" )

        klo = engine.rng.randint( 0, nsamp )

        p0 = wl[klo].allpars[:2]
        engine.calculateUnitRange()

        for k in range( 100 ) :
            wl[klo].allpars[:2] = p0
            engine.execute( wl[klo].id, lowL )
            p1 = wl[klo].allpars[:2]
            plt.plot( p1[0], p1[1], 'r.' )

        plt.plot( p0[0], p0[1], 'k*' )
        plt.show()

    def suite( cls ):
        return unittest.TestCase.suite( TestEngine2.__class__ )


class Plotter( object ) :

    def __init__( self, iter=0, kp=(0,1), figsize=(5,5) ) :
        self.iter = iter
        self.figsize = figsize
        self.k0 = kp[0]
        self.k1 = kp[1]
        self.col = ['k', 'r', 'g', 'b', 'm', 'y']
        self.sym = ['.', ',', 'o', '+', 'x', '*']

    def start( self, param=None ):
        """ start the plot. """
        plt.figure( 1, figsize=(5,5) )
        if param is not None :
            self.point( param, col=1, sym=2 )



    def point( self, param, col=0, sym=0 ) :
        """
        put a point at position param using color col.
        """
        k0 = self.k0
        k1 = self.k1
        cl = self.col[col] + self.sym[sym]
        plt.plot( [param[k0]], [param[k1]], cl )


    def move( self, param, ptry, col=0, sym=None ):
        """
        Move parameters at position param to ptry using color col.
        """
        k0 = self.k0
        k1 = self.k1
        cl = self.col[col] + '-'
        plt.plot( [param[k0], ptry[k0]], [param[k1], ptry[k1]], cl )

        if sym is not None :
            plt.text( ptry[k0], ptry[k1], "%d" % self.iter )
            self.point( ptry, col=col, sym=sym )
            self.iter += 1

    def stop( self, param=None, name=None ):
        """ Stop (show) the plot. """
        if param is not None :
            self.point( param, col=2, sym=2 )
        if name is not None :
            plt.title( name )

#        plt.xlabel( "Param %d" % self.k0 )
#        plt.ylabel( "Param %d" % self.k1 )
        plt.show()



if __name__ == '__main__':
    unittest.main( )





from __future__ import print_function

import numpy as numpy
import random
from astropy import units
import warnings

from .FixedModel import FixedModel
#from .Prior import Prior
#from .UniformPrior import UniformPrior
from . import Tools
from .Formatter import formatter as fmt
from .Tools import setAttribute as setatt

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2011 Do Kester, SRON (JAVA code)
#  *    2016 - 2023 Do Kester


class Model( FixedModel ):
    """
    Model implements the common parts of (compound) models.
    It is the last common anchestor of all Models.

    Models can be handled by the Fitter classes.

    A model consists of one or more instantiations of (base) models which
    are concatenated in a chain of models using various operations
    (+-*/). A special operation is the pipe (|). It works like a unix pipe,
    i.e. the output of the left-hand process in used as input of the
    right-hand process.

    Methods defined in BaseModel as eg. baseResult() are recursively called
    here as result(). They are the ones used in the fitters.

    The Model is the place where model-related items are kept, like parameters,
    stdevs.

    Model also implements a numerical derivation of partial to be
    used when partial is not given in the model definition itself. This same
    numerical derivation of partial is used in testPartial to indeed test
    whether the partial has been implemented properly.

    Example:
    --------
    >>> x = numpy.arange( 10 )
    >>> poly = PolynomialModel( 2 )             # quadratic model
    >>> poly.parameters = [3,2,1]               # set the parameters for the model
    >>> y = poly( x )                           # evaluate the model at x
    >>> p0 = poly[0]                            # 3: the first parameter
    >>>
    >>> # To make a compound model consisting of a gaussian and a constant background
    >>>
    >>> gauss = GaussModel( )                   # gaussian model
    >>> gauss += PolynomialModel( 0 )           # gaussian on a constant background
    >>> print( gauss.getNumberOfParameters( ) )
    >>> 4
    >>>
    >>> # Set limits to this model
    >>>
    >>> lolim = [0,-10,0,-5]                    # lower limits for the parameters
    >>> hilim = [10,10,2, 5]                    # high limits for parameters
    >>> gauss.setLimits( lolim, hilim )         # set limits. Does not work with all Fitters
    >>>
    >>> # Pipe a model; The order of operation matters.
    >>> # m5 = ( m1 | m2 ) + m3
    >>>
    >>> m1 = PolynomialModel( 1 )               # m1( x, p )
    >>> m2 = SineModel()                        # m2( x, q )
    >>> m3 = PolynomialModel( 0 )               # m3( x, r )
    >>> m4 = m1 | m2                            # m2( m1( x, p ), q )
    >>> m5 = m4 + m3                            # m2( m1( x, p ), q ) + m3( x, r )
    >>> print( m5.parameters )                  # [p, q, r]
    >>>
    >>> # Implicit brackets
    >>> # m5 = m1 | ( m2 + m3 )
    >>>
    >>> m1 = PolynomialModel( 1 )               # m1( x, p )
    >>> m2 = SineModel()                        # m2( x, q )
    >>> m3 = PolynomialModel( 0 )               # m3( x, r )
    >>> m4 = m2 + m3                            # m2( x, q ) + m3( x, r )
    >>> m5 = m1 | m4                            # m2( m1( x, p ), q ) + m3( m1( x, p ), r )
    >>> print( m5.parameters )                  # [p, q, r]

    Attributes
    ----------
    parameters : array_like
        parameters of the model
    stdevs : None or array_like
        standard deviations after a fit to the data
    xUnit : astropy.units or list of
        unit of the x-values (list of in case of more dimensions)
    yUnit : astropy.units
        unit of the y-values
    npars : int (read only)
        number of parameters in this model
    npchain : int (read only)
        identical to npars

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Author       Do Kester

    """

    NOP = 0
    ADD = 1
    SUB = 2
    MUL = 3
    DIV = 4
    PIP = 5

    #  *****CONSTRUCTOR*********************************************************
    def __init__( self, nparams=0, ndim=1, copy=None, params=None, **kwargs ):
        """
        Initializes the Model with all attributes set to None, except for
        the parammeters which are all initialized to 0.

        Parameters
        ----------
        nparams : int
            the number of parameters in this model
        ndim : int
            the dimensionality of the xdatas (default: 1)
        copy : Model
            model to be copied (default: None)
        params : array_like
            initial parameters of the model
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        super( Model, self ).__init__( nparams=nparams, ndim=ndim, copy=copy,
                            **kwargs )

        setatt( self, "_next", None )
        setatt( self, "_head", self )

        if params is None :
            params = numpy.zeros( nparams, dtype=float )
        else :
            params = Tools.toArray( params, dtype=float )
        setatt( self, "parameters", self.select( params ), type=float, islist=True )

        nparams = self.npbase                       # accounting len( fixed )
        setatt( self, "_npchain", nparams )
        setatt( self, "stdevs", None )

        # xUnit is by default a (list[ndim] of) scalars, unitless
        xUnit = units.Unit( 1.0 ) if ndim == 1 else [units.Unit( 1.0 )]*ndim
        setatt( self, "xUnit", xUnit )
        setatt( self, "yUnit", units.Unit( 1.0 ) )                  # scalar
        setatt( self, "_operation", self.NOP )

        if copy is None : return

        ## the head has been copied; append now the rest of the chain
        last = copy._next
        while last is not None:
            self.appendModel( last.isolateModel( 0 ), last._operation )
            last = last._next

        setatt( self, "_npchain", copy._npchain )
        if copy.parameters is not None :
            setatt( self, "parameters", copy.parameters.copy() )
        if copy.stdevs is not None :
            setatt( self, "stdevs", copy.stdevs.copy() )

        setatt( self, "xUnit", copy.xUnit )
        setatt( self, "yUnit", copy.yUnit )

    def copy( self ):
        """ Return a copy.  """
        return Model( copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes.

        Parameters
        ----------
        name :  string
            name of the attribute
        value :
            value of the attribute

        """

        if name in ['parameters', 'stdevs'] :
            if value is not None and Tools.length( value ) != self.npchain :
                raise ValueError( "%s: Nr of %s does not comply. expect %d; got %d" %
                                ( self.shortName(), name, self.npchain, Tools.length( value ) ) )
            setatt( self, name, value, type=float, islist=True, isnone=True )

        elif name in ['xUnit', 'yUnit'] :
            isl = ( name == 'xUnit' and self.ndim > 1 )
            setatt( self, name, value, type=units.core.UnitBase, islist=isl, isnone=True )

        elif name in ['pipePartial', 'pipeDeriv'] :
            setatt( self, name, value )

        else :
            super( Model, self ).__setattr__( name, value )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name == 'npars' or name == 'npchain' :
            return self._head._npchain
        elif name == 'ndout' :
            return 1
        elif name == 'lastndout' :
            last = self
            while last._next is not None : last = last._next
            return last.ndout

        return super( Model, self ).__getattr__( name )


    def chainLength( self ):
        """ Return length of the chain.  """
        last = self
        i = 0
        while last != None:
            last = last._next
            i += 1
        return i

    def isNullModel( self ) :
        """
        Return True if the model has no parameters (a NullModel).
        """
        return len( self.parameters ) == 0

    def isolateModel( self, k ):
        """
        Return a ( isolated ) copy of the k-th model in the chain.
        Fixed parameters and priors which might be present in the compound model
        will be lost in the isolated model.

        Parameters
        ----------
        k : int
            the model number ( head = 0 )

        Raises
        ------
        IndexError when the chain is shorter than k.

        """
        last = self
        i = 0
        np = 0
        while last is not None:
            if i == k :
                next = last._next               #  save

                setatt( last, "_next", None )   #  isolate
                head = last._head               #  save
                setatt( last, "_head", last )   #  isolate
                mdl = last.copy()               #  copy
                setatt( last, "_next", next )   #  restore
                setatt( last, "_head", head )
                setatt( mdl, "_operation", self.NOP )

                n2 = np + last.npbase
                # hard overwrite of the base parameters (over the chain parameters)
                setatt( mdl, "parameters", self._head.parameters[np:n2] )

                if self.stdevs is not None :
                    setatt( mdl, "stdevs", self._head.stdevs[np:n2] )
                else :
                    setatt( mdl, "stdevs", None )
                setatt( mdl, "_npchain", mdl.npbase )      # only extract one model
                return mdl

            np += last.npbase
            last = last._next
            i += 1
        raise IndexError( "There are only %d models in this compound model" % i )

    #  *************************************************************************
    def addModel( self, model ):
        """
        Make a compound model by concatinating/adding model to this.

        The final result is the sum of the individual results.

        The compound model is implemented as a chain of Models.
        Each of these base models contain the attributes ( parameters, limits etc. )
        and when needed these attributes are taken from there, or stored there.

        The operation (addition in this case) is always with the total result of the
        existing chain. For the use of "brackets" in a chain use BracketModel.

        Parameters
        ----------
        model : Model
            model to be added to

        """
        self.appendModel( model, self.ADD )

    def subtractModel( self, model ):
        """
        Make a compound model by concatinating/subtracting a model from this.

        The final result is the difference of the models.

        Parameters
        ----------
        model : Model
            model to be subtracted from

        """
        self.appendModel( model, self.SUB )

    def multiplyModel( self, model ):
        """
        Make a compound model by concatinating/multiplying a model with this.

        The final result is the product of the models.

        Parameters
        ----------
        model : Model
            model to be multiplied by

        """
        self.appendModel( model, self.MUL )

    def divideModel( self, model ):
        """
        Make a compound model by concatinating/dividing by a model.

        The final result is the division of the models.

        Parameters
        ----------
        model : Model
            model to be divided by

        """
        self.appendModel( model, self.DIV )

    def pipeModel( self, model ):
        """
        Make a compound model by piping the result into the next.

        Parameters
        ----------
        model : Model
            model to pipe into

        """
        self.appendModel( model, self.PIP )

    def appendModel( self, model, operation ):
        """
        Append a model to the present chain using a operation.

        Parameters
        ----------
        model : Model
            the model to be appended
        operation : int
            operation index

        Raises
        ------
        ValueError when a model of a different dimensionality is offered.
        """
        if self.isDynamic() and model.isDynamic():
            raise ValueError( "Only one Dynamic model in a chain" )
        if operation != self.PIP :
            opname = ["", "+", "-", "*", "/", "|" ]
            if self.ndim != model.ndim:
                raise ValueError( "Trying to combine incompatible model dimensions: %d %s %d"%
                                (self.ndim, opname[operation], model.ndim) )
        else :
            if self.lastndout != model.ndim :
                raise ValueError( "Trying to pipe incompatible models:  %d | %d"%
                                (self.lastndout, model.ndim) )
            model.selectPipe( self.ndim, model.ndim, model.ndout )

        if model._next is not None :
            model = Brackets( model )     # provide brackets if model is a chain

        last = self
        while last._next != None:
            last = last._next
        setatt( last, "_next",  model )
        setatt( model, "_operation", operation )
        while last._next != None:
            last = last._next
            setatt( last, "_head", self._head )

        setatt( self, "_npchain", len( self.parameters ) + len( model.parameters ) )

        setatt( self, "parameters", self._optAppend( self.parameters, model.parameters ) )

        # Erase the model's attributes; not needed anymore
        setatt( model, "parameters", None )

        return

    def _optAppend( self, x, y ) :
        if y is not None :
            if x is None : return y
            else : return numpy.append( x, y )
        else : return x

    #  *****CHECK**************************************************************
    def correctParameters( self, params ):
        """
        Check parameters for non-zero and positivity

        Parameters
        ----------
        params : array_like
            parameters for the model.

        """
        newpar = numpy.asarray( [], dtype=float )

        pp = self._recursiveCorrect( params, newpar )
        return pp

    def _recursiveCorrect( self, params, newpar ) :

        np = self.npbase
        newpar = numpy.append( newpar,  super( Model, self ).checkParameter( params[:np] ) )
        model = self._next
        if model is None :
            return newpar

        newpar = model._recursiveCorrect( params[np:], newpar )
        return newpar

    #  *****RESULT**************************************************************
    def result( self, xdata, param=None ):
        """
        Return the result of the model as applied to an array of input data.

        Parameters
        ----------
        xdata : array_like
            input data
        param : array_like
            parameters for the model. Default parameters from the Model

        """
        if param is None :
            param = self.parameters

        res = None
        xdata = Tools.toArray( xdata )
        return self._recursiveResult( xdata, param, res )

    def _recursiveResult( self, xdata, param, res ) :

        np = self.npbase

        if not self._operation == self.PIP :
            nextres = super( Model, self ).result( xdata, param[:np] )
        else :
            nextres = None
        res = self.operate( res, param[:np], nextres )
        model = self._next
        if model is None :
            return res

        res = model._recursiveResult( xdata, param[np:], res )
        return res

    def operate( self, res, pars, next ):
        if res is None or self._operation == self.NOP: # first one
            res = next
        elif self._operation == self.ADD:                 # NOP & ADD
            res = numpy.add( res, next )
        elif self._operation == self.SUB:
            res = numpy.subtract( res, next )
        elif self._operation == self.MUL:
            res = numpy.multiply( res, next )
        elif self._operation == self.DIV:
            res = numpy.divide( res, next )
        elif self._operation == self.PIP:
            res = super( Model, self ).result( res, pars )

        return res

    #  *****DERIVATIVE*********************************************************
    def derivative( self, xdata, param, useNum=False ):
        """
        Return the derivatives (df/dx) of the model at the inputs

        Parameters
        ----------
        xdata : array_like
            an input vector or array
        param : array_like
            parameters for the model
        useNum : bool
            if true, numeric derivatives are used.

        """
        result = None
        xdata = Tools.toArray( xdata )

        df = self._recursiveDerivative( xdata, param, result, 0,
                    useNum=useNum )
        return df

    def _recursiveDerivative( self, xdata, param, result, df, useNum=False ):
        """
        Workhorse for derivative.

        Implements the optional concatenation of models via recursion

        """
        np = self.npbase
        par = param[:np]
        nextres = None

#        if self.npmax > 0:            #  the base model has no parameters at all: skip
        xd = xdata if self._operation != self.PIP else result
        nextdf = ( super( ).numDerivative( xd, par ) if useNum else 
                       super( ).derivative( xd, par ) )

#        else :
#            nextdf = numpy.zeros_like( xdata, dtype=float )

#        print( "Model2  ", self.shortName(), df.shape, nextdf.shape, self._operation )

        isnotlist = self.ndim == 1 or self.ndout == 1
        if self._operation == self.NOP :
            df = nextdf

        elif self._operation == self.ADD :
            df = ( df + nextdf if isnotlist else 
                 [ d + n for d,n in zip( df, nextdf )] )

        elif self._operation == self.SUB :
            df = df - nextdf

        elif self._operation == self.MUL :
            nextres = super( Model, self ).result( xdata, par )
            df = df * nextres + nextdf * result

        elif self._operation == self.DIV :
            nextres = super( Model, self ).result( xdata, par )
            df = ( df * nextres - nextdf * result ) / ( nextres * nextres )

        elif self._operation == self.PIP :
#            nextres = 'dummy'           ## not needed, but needs something
            df = self.pipeDeriv( df, nextdf )
        else :
            raise ValueError( "Unknown operation: %d" & self.PIP )

        if self._next is None :
            return df
        if nextres is None:
            nextres = super( Model, self ).result( xdata, par )
        result = self.operate( result, par, nextres )

        #  append the dfs of the _next model
        model = self._next
        return model._recursiveDerivative( xdata, param[np:], result, df,
                    useNum=useNum )

    #  *****PARTIAL*************************************************************
    def partial( self, xdata, param, useNum=False ):
        """
        Return the partial derivatives of the model at the inputs

        Parameters
        ----------
        xdata : array_like
            an input vector or array
        param : array_like
            parameters for the model
        useNum : bool
            if true, numeric partials are used.

        """
#        print( "Partial ", self.shortName( ), self._operation )

        result = None
        partial = None
        xdata = Tools.toArray( xdata )
        partial = self._recursivePartial( xdata, param, 0, result,
                            partial, useNum=useNum )
        return partial

    def _recursivePartial( self, xdata, param, at, result, partial, useNum=False ):
        """
        Workhorse for partial.

        Implements the optional concatenation of models via recursion

        """
        np = self.npbase
        par = param[at:at+np]
        nextres = None

        if np > 0 :
            xd = xdata if self._operation != self.PIP else result

            nextpartial = ( super( ).numPartial( xd, par ) if useNum else
                            super( ).partial( xd, par ) )

        else :
            inlen = Tools.length( xdata )
            nextpartial = numpy.ndarray( (inlen,0), dtype=float )

        if self._operation == self.SUB :
            nextpartial = numpy.negative( nextpartial )

        elif self._operation == self.MUL :
            nextres = super( Model, self ).result( xdata, par )
            partial = numpy.multiply( partial.transpose(), nextres ).transpose()
            nextpartial = numpy.multiply( nextpartial.transpose(), result ).transpose()

        elif self._operation == self.DIV :
            nextres = super( Model, self ).result( xdata, par )
            partial = numpy.divide( partial.transpose(), nextres ).transpose()
            invres = - result / ( nextres * nextres )
            nextpartial = numpy.multiply( nextpartial.transpose(), invres ).transpose()

        elif self._operation == self.PIP :

            # Need derivative: dHdG
            dHdG = ( super( ).numDerivative( result, par ) if useNum else 
                     super( ).derivative( result, par ) )

            partial = self.pipePartial( partial, dHdG )

        if not isinstance( partial, list ) :
            partial = ( nextpartial if partial is None
                    else numpy.append( partial, nextpartial, axis=1 ) )
        else :
            for k in range( len( partial ) ) :
                partial[k] = numpy.append( partial[k], nextpartial[k], axis=1 )

        model = self._next
        if model is None:
            return partial
        if nextres is None:
            nextres = super( Model, self ).result( xdata, par )

        result = self.operate( result, param[at:], nextres )
        #  append the partials of the _next model
        at += np
        return model._recursivePartial( xdata, param, at, result, partial,
                    useNum=useNum )


    def selectPipe( self,  ndim, ninter, ndout ) :

        # Models G pars p and H pars q

        # F(x:pq) ==>  H(G(x:p):q)            G(x:p) | H(*:q)   
        # dF/dp   ==>  dH/dG * dG/dp          H.derivative(G,p) * G.partial(x,q)
        # dF/dq   ==>  dH(G(x:p):q) / dq      G.partial(H,q)

        # G.ndout mustbe H.ndim
        # partial <== G
        # dfdx    <== H


        if ( ndim == 1 and ninter == 1 and ndout == 1 ) :
            self.pipePartial = self.pipe_0
            self.pipeDeriv   = self.pipe_4
        elif ( ndim > 1 and ninter == 1 and ndout == 1 ) :
            self.pipePartial =  self.pipe_0
            self.pipeDeriv   = self.pipe_0
       
        elif ( ndim == 1 and ninter > 1 and ndout > 1 ) :
            self.pipePartial = self.pipe_1
            self.pipeDeriv   = self.pipe_5

        elif ( ndim > 1 and ninter > 1 and ndout > 1 ) :
            self.pipePartial = self.pipe_1
            self.pipeDeriv   = self.pipe_1

        elif ( ndim == 1 and ninter > 1 and ndout == 1 ) :
            self.pipePartial = self.pipe_2
            self.pipeDeriv   = self.pipe_6

        elif ( ndim == 1 and ninter == 1 and ndout > 1 ) :
            self.pipePartial = self.pipe_3
            self.pipeDeriv   = self.pipe_7

        elif ( ndim > 1 and ninter > 1 and ndout == 1 ) :
            self.pipePartial = self.pipe_2
            self.pipeDeriv   = self.pipe_8

        elif ( ndim > 1 and ninter == 1 and ndout > 1 ) :
            self.pipePartial = self.pipe_3
            self.pipeDeriv   = self.pipe_9

        return


    ########################################################################
    # Next follow a number of pipe operations for partials.
    #    F( x:p,q ) = H( G( x:p ), q ) or
    #                 G( x:p ) | H( x:q )
    #    dF/dp = dH/dG * dG/dp
    #          = H.derive * G.partial
    #    dF/dx = dH/dG * dG/dx
    #          = H.derive * G.derivative
    # Different cases defined by the dimensions of of input partial
    # where N : number of datapoints x
    #       I : dimension of x (inputs)
    #       K : dimension of G (intermediate: out for G; in for H)
    #       O : dimension of F (outputs)
    #       P : number of parameters
    ########################################################################

    def pipe_0( self, dGd, dHdG ) :
        """
        ninter == 1 and ndout == 1
        Return partial in the form of [N,P]
        
        Parameters
        ----------
        dGd:  array of form [N,P]
            Either partial dGdp or derivative dGdx  
        dHdG: array of form [N]
            Derivative of H to G  
        """
        return numpy.multiply( dGd.transpose(), dHdG ).transpose()

    def pipe_1( self, dGd, dHdG ) :
        """
        ninter > 1 and ndout > 1
        Return partial in the form [O][N,P]
        
        Parameters
        ----------
        dGd:  array of form [K][N,P]
            Either partial dGdp or derivative dGdx  
        dHdG: array of form [O][N,K]
            Derivative of H to G  
        """
        parr = numpy.asarray( dGd )

        part = []
        for dd in dHdG :
            pp = 0
            for k in range( len( dGd ) ) :
                pp += numpy.multiply( dGd[k].T, dd[:,k] )
            part += [pp.T]

        return part

    def pipe_2( self, dGd, dHdG ) :
        """
        ndim == 1 and ninter > 1 and ndout == 1
        Return partial in the form of [N,P]
        
        Parameters
        ----------
        dGd:  array of form [K][N,P]
            Either partial dGdp or derivative dGdx  
        dHdG: array of form [N,K]
            Derivative of H to G  
        """
        parr = numpy.asarray( dGd )
        part = numpy.zeros_like( dGd[0] )

        for k in range( dGd[0].shape[1] ) :
            part[:,k] = numpy.sum( dHdG.T * parr[:,:,k], axis=0 )
            
        return part

    def pipe_3( self, dGd, dHdG ) :
        """
        ndim == 1 and ninter = 1 and ndout > 1
        Return partial in the form of [O][NP]
        
        Parameters
        ----------
        dGd:  array of form [N,P]
            Either partial dGdp or derivative dGdx  
        dHdG: array of form [N,0]
            Derivative of H to G  
        """
        part = []
        for k in range( dHdG.shape[1] ) :
            part += [ numpy.multiply( dGd.transpose(), dHdG[:,k] ).transpose() ]

        return part

    def pipe_4( self, dGdx, dHdG ) :
        """
        ndim == 0 and ninter == 1 and ndout == 1
        Return partial in the form of [N]
        
        Parameters
        ----------
        dGdx: array of form [N]
            Derivative dGdx  
        dHdG: array of form [N]
            Derivative of H to G  
        """
        return dGdx * dHdG

    def pipe_5( self, dGdx, dHdG ) :
        """
        ndim == 1 and ninter > 1 and ndout > 1
        Return derivative in the form of [N,O]
        
        Parameters
        ----------
        dGdx:  array of form [N,K]
            Either partial dGdp or derivative dGdx  
        dHdG: array of form [O][N,K]
            Derivative of H to G  
        """
        darr = numpy.asarray( dHdG )
        part = numpy.zeros_like( darr[:,:,0] ).T

        for k in range( part.shape[1] ) :
            part[:,k] = numpy.sum( numpy.multiply( darr[k,:,:], dGdx ), axis=1 )

        return part

    def pipe_6( self, dGdx, dHdG ) :
        """
        ndim == 1 and ninter > 1 and ndout == 1
        Return derivative in the form of [N]
        
        Parameters
        ----------
        dGdx:  array of form [N,K]
            Either partial dGdp or derivative dGdx  
        dHdG: array of form [N,K]
            Derivative of H to G  
        """
        part = numpy.sum( numpy.multiply( dHdG, dGdx ), axis=1 )
        return part

    def pipe_7( self, dGdx, dHdG ) :
        """
        ndim == 1 and ninter = 1 and ndout > 1
        Return derivative in the form of [N,O]
        
        Parameters
        ----------
        dGdx:  array of form [N,O]
            Either partial dGdp or derivative dGdx  
        dHdG: array of form [N]
            Derivative of H to G  
        """
        part = numpy.multiply( dHdG.T, dGdx.T ).T

        return part

    def pipe_8( self, dGdx, dHdG ) :
        """
        ndim > 1 and ninter > 1 and ndout == 1
        Return derivative in the form of [N,I]
        
        Parameters
        ----------
        dGdx:  array of form [K][N,I]
            Either partial dGdp or derivative dGdx  
        dHdG: array of form [N,K]
            Derivative of H to G  
        """
        darr = numpy.asarray( dGdx )

        part = numpy.zeros_like( dGdx[0] )
        for k in range( part.shape[1] ) :
            part[:,k] = numpy.sum( numpy.multiply( dHdG.T, darr[:,:,k] ), axis=0 )

        return part

    def pipe_9( self, dGdx, dHdG ) :
        """
        ndim > 1 and ninter == 1 and ndout > 1
        Return derivative in the form of [O][N,I]
        
        Parameters
        ----------
        dGdx:  array of form [N,I]
            Either partial dGdp or derivative dGdx  
        dHdG: array of form [N,O]
            Derivative of H to G  
        """
        part = []
        for k in range( dHdG.shape[1] ) :
            part += [numpy.multiply( dGdx.T, dHdG[:,k] ).T]

        return part


    #  *****TOSTRING***********************************************************
    def __str__( self ):
        """ Returns a string representation of the model.  """
        return self._toString( "" )

    def _toString( self, indent, npars=0 ) :
        opname = [" null\n", " +\n", " -\n", " *\n", " /\n", " |\n" ]
        np = self.npbase

        if self._next is None :
            return super( Model, self )._toString( npars=npars )

        return ( super( Model, self )._toString( npars=npars ) +
                 opname[self._next._operation] +
                 indent + self._next._toString( indent, npars=npars+np ) )

    def shortName( self ) :
        """
        Return a short version the string representation: upto first non-letter.
        """
        opname = [" null\n", " + ", " - ", " * ", " / ", " | " ]
        if self._next is None :
            return super( Model, self ).shortName()

        return ( super( Model, self ).shortName( ) +
                 opname[self._next._operation] +
                 self._next.shortName( ) )


    #  *****GET/SET*************************************************************
    def getNumberOfParameters( self ):
        """ Returns the number of parameters of the ( compound ) model.  """
        return len( self._head.parameters )

    #  *****NUMERIC DERIVATIVE*************************************************
    def numDerivative( self, xdata, param ):
        """
        Returns numerical derivatives (df/dx) of the model function.

        Parameters
        ----------
        xdata : array_like
            input data
        param : array_like
            a parameters vector

        """
        return self.derivative( xdata, param, useNum=True )

    #  *****NUMERIC PARTIAL****************************************************
    def numPartial( self, xdata, param ):
        """
        Returns numerical partial derivatives of the model function.

        Parameters
        ----------
        xdata : array_like
            input data
        param : array_like
            a parameters vector

        """
        return self.partial( xdata, param, useNum=True )

    def isDynamic( self ) :
        """
        Return whether the model can change the number of parameters dynamically.
        """
        if super( Model, self ).isDynamic() :
            return True
        elif self._next is not None :
            return self._next.isDynamic()

        return False

    #  *****PRIOR***************************************************************
    def hasPriors( self, isBound=True ) :
        """
        Return True when the model has priors for all its parameters.

        Parameters
        ----------
        isBound : bool
            Also check if the prior is bound.
        """
        hasp = True
        mdl = self
        while mdl is not None and hasp :
            hasp = hasp and super( Model, self ).hasPriors( isBound=isBound )
            mdl = mdl._next
        return hasp

    def getPrior( self, k ):
        """
        Return the prior of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        Raises
        ------
        IndexError when k is larger than the number of parameters.

        """
        np = self.npbase
        if k < np:
            return super( Model, self  ).getPrior( k )
        elif self._next != None:
            return self._next.getPrior( k - np )
        else:
            raise IndexError( "The (compound) model does not have %d parameters"%( k + 1 ) )

    def setPrior( self, k, prior=None, **kwargs ):
        """
        Set the prior for the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.
        prior : Prior
            prior for parameter k
        kwargs : keyword arguments
            attributes to be passed to the prior

        Raises
        ------
        IndexError when k is larger than the number of parameters.

        """
        np = self.npbase
        if k < np:
            super( Model, self  ).setPrior( k, prior=prior, **kwargs )
        elif self._next != None:
            self._next.setPrior( k - np, prior=prior, **kwargs )
        else:
            raise IndexError( "The (compound) model does not have %d parameters"%( k + 1 ) )


    #  ***PARAMETER NAME *******************************************************
    def getParameterName( self, k ):
        """
        Return the name of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        Raises
        ------
        IndexError when k is larger than the number of parameters.

        """
        np = self.npbase
        if k < np:
            return super( Model, self ).getParameterName( k )
        elif self._next != None:
            return self._next.getParameterName( k - np )
        else:
            raise IndexError( "The (compound) model does not have " + str( k + 1 ) +
                              " parameters." )

    def getParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        Raise
        -----
        IndexError when k is > number of parameters

        """
        np = self.npbase
        if k < np:
            return super( Model, self).getParameterUnit( k )
        elif self._next != None:
            return self._next.getParameterUnit( k - np )
        else:
            raise IndexError( "The (compound) model does not have " + str( k + 1 ) +
                              " parameters." )

    def getIntegralUnit( self ):
        """ Return the unit of the integral of the model over x. """

        unit = self.yUnit
        if isinstance( self.xUnit, list ) :
            for u in self.xUnit : unit *= u
        else : unit *= self.xUnit
        return unit


    #  *****LIMITS**************************************************************
    def setLimits( self, lowLimits=None, highLimits=None ):
        """
        Sets the limits for the parameters of the compound model.

        1. It is valid to insert for either parameter a None value
        indicating no lower/upper limits.
        2. When a lowerlimit >= upperlimit no limits are enforced.
        It only works in *Fitter classes which support it.

        Parameters
        ----------
        lowLimits : array_like
            lower limits on the parameters
        highLimits : array_like
            upper limits on the parameters

        """
        ml = 0
        if lowLimits is not None :
            lowLimits = Tools.toArray( lowLimits )
            nl = len( lowLimits )
            ml = max( ml, nl )
        if highLimits is not None :
            highLimits = Tools.toArray( highLimits )
            nh = len( highLimits )
            ml = max( ml, nh )
        if ml == 0 : return

        if ml > self.npchain :
            warnings.warn( "More limits given than parameters present: %d < %d" %
                            ( ml, self.npchain ) )

        for k in range( ml ) :
            lo = None if k >= nl else lowLimits[k]
            hi = None if k >= nh else highLimits[k]
            self.setPrior( k, limits=[lo,hi] )

    def getLimits( self ) :
        """
        Return the limits stored in the priors

        Returns
        -------
        limits : tuple of 2 array-like or of 2 None (if `self.priors` is None)
            (lowlimits, highlimits)

        lolim = []
        hilim = []
        mdl = self
        while mdl is not None :
            if not super( Model, mdl ).hasLimits( ) :
                return [None,None]
            lolim += [p.lowLimit for p in mdl.priors]
            hilim += [p.highLimit for p in mdl.priors]

            mdl = mdl._next
        return (lolim, hilim)
        """

        lolim = [self.getPrior(k).lowLimit for k in range( self.npchain )]
        hilim = [self.getPrior(k).highLimit for k in range( self.npchain )]
        return (lolim, hilim)




    #  *************************************************************************
    def hasLimits( self, fitindex=None ):
        """
        Return true if limits has been set for this model.

        Parameters
        ----------
        fitindex    list of indices to be fitted.

        """
        haslim = True
        mdl = self
        while mdl is not None :
            haslim = haslim and super( Model, mdl ).hasLimits( fitindex=fitindex )
            if not haslim :
                return haslim
            if fitindex is not None :
                q = numpy.where( fitindex >= mdl.npbase )
                fitindex = fitindex[q] - mdl.npbase
            mdl = mdl._next
        return haslim


    #  ****** UNIT <--> DOMAIN ********************************************
    def unit2Domain( self, uvalue, kpar=None ):
        """
        Convert a value in [0,1] to one inside the limits of the parameter.

        Parameters
        ----------
        uvalue : (list of) float
            value in [0,1]
        kpar : int
            index of the parameter

        """
        if kpar is not None :
            return self.getPrior( kpar ).unit2Domain( uvalue )

        pgen = self.nextPrior(  )
        dval = numpy.fromiter( ( next( pgen ).unit2Domain( uv ) for uv in uvalue ), float )
        try :
            pgen.close()
        except :
            pass
        return dval


    def domain2Unit( self, dvalue, kpar=None ):
        """
        Convert a value within the domain of the parameter to one in [0,1].

        Parameters
        ----------
        dvalue : (list of) float
            value of parameter
        kpar : int
            index of the parameter

        """
        if kpar is not None :
            return self.getPrior( kpar ).domain2Unit( dvalue )

        pgen = self.nextPrior()
        uval = numpy.fromiter( ( next( pgen ).domain2Unit( dv ) for dv in dvalue ), float )
        try :
            pgen.close()
        except :
            pass
        return uval

    def partialDomain2Unit( self, dvalue ):
        """
        Return a the derivate of the domain2Unit function to dval.

        Parameters
        ----------
        dvalue : (list of) float
           parameter array

        """
        pgen = self.nextPrior(  )
        part = numpy.fromiter( ( next( pgen ).partialDomain2Unit( dv ) for dv in dvalue ), float )
        try :
            pgen.close()
        except :
            pass
        return part

    def nextPrior( self ) :
        mdl = self
        k = 0
        while True :
            try :
                yield mdl.priors[k]
            except :
                yield mdl.priors[-1]
            k += 1
            if k >= mdl.npbase :
                mdl = mdl._next
                k = 0


    #  *****SOME DUMMY METHODS TO ALLOW LINEAR MODELS IN NONLIN FITTERS********
    def isMixed( self ):
        """ Return false.  """
        return False

    def getLinearIndex( self ):
        """ Return null.  """
        return None

    #  ***** PYTHON INTERFACES ****************************************************
    def __getitem__( self, i ):
        """
        Return the i-th parameter.

        >>> p_i = model[i]

        Parameters
        ----------
        i : int
            index for the parameter.

        """
        return self.parameters[i]

    def __call__( self, x ):
        """
        Return the result of the model.

        >>> y = model( x )
        is equivalent to
        >>> y = model.result( x, model.parameters )

        Parameters
        ----------
        x : (list of) float
            apply the model to x (as input)

        """
        return self.result( x, self.parameters )

    def __iadd__( self, model ):
        """
        Method for making compound models using += operator.

        >>> model1 += model2

        Parameters
        ----------
        model : Model
            a model to add to self (the existing chain)

        """
        self.addModel( model )
        return self

    def __add__( self, model ):
        """
        Method for making compound models using + operator.

        >>> model1 = model2 + model3

        Parameters
        ----------
        model : Model
            a copy of model to add to a copy of self (the existing chain)

        """
        return self.copy().__iadd__( model.copy() )

    def __isub__( self, model ):
        """
        Method for making compound models using -= operator.

        >>> model1 -= model2

        Parameters
        ----------
        model : Model
            a model to subtract from self (the existing chain)

        """
        self.subtractModel( model )
        return self

    def __sub__( self, model ):
        """
        Method for making compound models using - operator.

        >>> model1 = model2 - model3

        Parameters
        ----------
        model : Model
            a copy of model to subtract from a copy of self (the existing chain)

        """
        return self.copy().__isub__( model.copy() )

    def __imul__( self, model ):
        """
        Method for making compound models using *= operator.

        >>> model1 *= model2

        Parameters
        ----------
        model : Model
            a model to multiply with self (the existing chain)

        """
        self.multiplyModel( model )
        return self

    def __mul__( self, model ):
        """
        Method for making compound models using * operator.

        >>> model1 = model2 * model3

        Parameters
        ----------
        model : Model
            a copy of model to multiply with a copy of self (the existing chain)

        """
        return self.copy().__imul__( model.copy() )

    def __itruediv__( self, model ):
        """
        Method for making compound models using /= operator.

        >>> model1 /= model2

        Parameters
        ----------
        model : Model
            a model to divide self with (the existing chain)

        """
        self.divideModel( model )
        return self

    def __truediv__( self, model ):
        """
        Method for making compound models using / operator.

        >>> model1 = model2 / model3

        Parameters
        ----------
        model : Model
            a copy of model to divide a copy of self with (the existing chain)

        """
        return self.copy().__itruediv__( model.copy() )

    def __ior__( self, model ):
        """
        Method for making compound models using |= operator.

        >>> model1 |= model2

        Use the results of model1 as input for model2.

        Parameters
        ----------
        model : Model
            a model to pipe the previous results through

        """
        self.pipeModel( model )
        return self

    def __or__( self, model ):
        """
        Method for making compound models using | (pipe) operator.

        >>> model1 = model2 | model3

        Parameters
        ----------
        model : Model
            a copy of model to pipe the results of a copy of self (the existing chain)

        """
        return self.copy().__ior__( model.copy() )

    #  *************************************************************************
    def testPartial( self, xdata, params, silent=True ):
        """
        A test routine to check the calculation of the partial derivatives.

        It is compared to a numerical calculation.

        Parameters
        ----------
        xdata : (list of) float
            values of the independent variable
        params : list of floats
            parameters for the model
        silent : bool
            if false print outputs

        Return
        ------
        The number of large deviations.

        """
        xdata  = Tools.toArray( xdata, ndim=self.ndim )
        params = Tools.toArray( params )
        sz = xdata.shape
        res = self.result( xdata, params )

        random.seed( 13010804 )

        snum = self.strictNumericDerivative( xdata, params )
        try :
            df = self.derivative( xdata, params )
            if df is None :
                raise ValueError
        except :
            df = snum
            if not silent :
                print( "The model has no derivatives df/dx" )

#        print( df )
#        print( snum )


        snmp = self.strictNumericPartial( xdata, params )
        try :
            partial = self.partial( xdata, params )
        except :
            partial = snmp
            if not silent :
                print( "The model has no partials df/dp" )


        ndout = self.lastndout

        kerr = 0
        lrang = range( sz[0] )
        if sz[0] > 10 :
            lrang = numpy.sort( random.sample( lrang, 10 ) )


        spaces = "          "
        for k in lrang :
            if not silent :
                print( "xdata[%2d]" % k, end='' )
                if self.ndim == 1 :
                    print( "   %8.3f"%(xdata[k]), end='' )
                else :
                    for i in range( sz[1] ) :
                        print( "   %8.3f"%(xdata[k,i]), end='' )

                print( "       result  ",  fmt( res[k] ) )
                print( "    par         value      partial      numeric" )

            if isinstance( df, list ) :
                df   = numpy.asarray( df )
                snum = numpy.asarray( snum )

            ms = numpy.maximum( numpy.abs( df + snum ), 1.0 )
            q = numpy.where( numpy.abs( df - snum ) > 0.001 * ms )
            lq = len( q[0] )
            kerr += lq

            if self.ndim == ndout == 1:
                if not silent :
                    print( "   dfdx               %12.5f %12.5f" % ( df[k], snum[k] ) )
            elif ( self.ndim > 1 ) and ( ndout > 1 ) : 
                for i in range( self.ndim ) :
                    ksi = 10 if i < 10 else 9
                    for j in range( ndout ) :
                        ks = ksi if j < 10 else ksi - 1
                        if not silent :
                            print( "  df%idx%i   %s %12.5f %12.5f" %
                                ( i, j, spaces[:ks], df[j,k,i], snum[j,k,i] ) )
            else :
                for i in range( max( self.ndim, ndout ) ) :
                    ks = 10 if i < 10 else 9
                    if not silent :
                        print( "  dfdx%i   %s  %12.5f %12.5f" %
                           ( i, spaces[:ks], df[k,i], snum[k,i] ) )


            if isinstance( partial, list ) :
                partial = numpy.asarray( partial )
                snmp    = numpy.asarray( snmp )

            ms = numpy.maximum( numpy.abs( partial + snmp ), 1.0 )
            q = numpy.where( numpy.abs( partial - snmp ) > 0.001 * ms )
            lq = len( q[0] )
            kerr += lq

            for i in range( self.npars ) :
                if ( lq >= 1 and i in q[-1] ) or random.random() < 5.0 / self.npars :
                    ks = 4 if i < 10 else 3 if i < 100 else 2
                        
                    if ndout == 1 : 
                        if not silent :
                            print( "  dfdp%d %s %8.3f %12.5f %12.5f"%
                            (i, spaces[:ks], params[i], partial[k,i], snmp[k,i]) )
                    else : 
                        for j in range( ndout ) :
                            if not silent :
                                print( " df%ddp%d %s %8.3f %12.5f %12.5f"%
                                (j, i, spaces[:ks], params[i], partial[j,k,i], snmp[j,k,i]) )
                    
        return kerr

    def strictNumericPartial( self, xdata, params, parlist=None ) :
        """
        Strictly numeric calculation of partials to params.

        For compound models it is different from numPartial and numDerivative.

        Parameters
        ----------
        xdata : float
            single xdata point (possibly multidimensional)
        params : array-like
            parameters
        kpar : None or int or list of int
            int  : return derivative to parameter kpar.
        """
        nxdata = Tools.length( xdata )
        if self.lastndout <= 1 :
            partial = numpy.zeros( ( nxdata, self.npars ), dtype=float  )
            assignPart = self.assignDF1
        else :
            partial = [ numpy.zeros( ( nxdata, self.npars ), dtype=float  ) 
                        for k in range( self.lastndout ) ]
            assignPart = self.assignDF2 

        dpg = Tools.makeNext( self.deltaP, 0 )
        
        par = params.copy()
        if parlist is None :
            parlist = numpy.arange( self.npars )
        
        for i, k in enumerate( parlist ) :
            dp = next( dpg )   
            par[k] += 0.5 * dp
            r1 = self.result( xdata, par )
            par[k] -= dp   
            r2 = self.result( xdata, par )

            assignPart( partial, i, ( r1 - r2 ) / dp )
            par[k] = params[k]
        try :
            dpg.close()
        except :
            pass
        return partial

    def assignDF1( self, partial, i, dpi ) :
        partial[:,i] = dpi

    def assignDF2( self, partial, i, dpi ) :

#        print( "assignDF2  ", len( partial ), partial[0].shape, dpi.shape )
        for k in range( self.lastndout ) :
            partial[k][:,i] = dpi[:,k]


    def strictNumericDerivative( self, xdata, param ) :
        """
        Strictly numeric calculation of derivative.

        For compound models it is different from numPartial and numDerivative.

            ## More dimensions in x

        Parameters
        ----------
        xdata : float
            single xdata point (possibly multidimensional)
        param : array-like
            parameters
        """

        dx = self.deltaP[0]

        if self.ndim == 1:
            r1 = self.result( xdata+dx, param )
            r2 = self.result( xdata-dx, param )
            return ( r1 - r2 ) / ( 2 * dx )

        ## More dimensions in xdata
        if self.lastndout == 1 :
            df = numpy.zeros_like( xdata )
            assignDF = self.assignDF1
        else :
            df = [ numpy.zeros_like( xdata ) for k in range( self.lastndout ) ]
            assignDF = self.assignDF2

        for i in range( self.ndim ) :
            x = xdata.copy()
            x[:,i] += dx
            r1 = self.result( x, param )
            x[:,i] -= 2 * dx
            r2 = self.result( x, param )

            assignDF( df, i, ( r1 - r2 ) / ( 2 * dx ) )

        return df


##### End Model #########################################################


class Brackets( Model ):
    """
    Brackets is only for use in Model. Use BracketModel for independent uses.

    """

    #  *************************************************************************
    def __init__( self, model, copy=None, **kwargs ):

        super( Brackets, self ).__init__( model.npchain, ndim=model.ndim,
                                copy=copy, **kwargs )
        setatt( self, "model", model, type=Model )
        setatt( self, "parameters", model.parameters )

        next = model
        deep = 1
        while next is not None :
            if isinstance( next, Brackets ) :
                deep += 1
            next = next._next
        setatt( self, "deep", deep )
        setatt( self, "parNames", [] )

    def copy( self ):

        return Brackets( self.model.copy(), copy=self )


    #  *****Brackets RESULT**************************************************************
    def baseResult( self, xdata, param ):
        """
        Returns the result calculated at the xdatas.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        return self.model.result( xdata, param )

    #  *****Brackets PARTIAL*************************************************************
    def basePartial( self, xdata, param, parlist=None ):
        """
        Returns the partial derivatives calculated at the xdatas.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            Not in use

        """
        return self.model.partial( xdata, param )

    #  *****Brackets DERIVATIVE***********************************************************
    def baseDerivative( self, xdata, param ):
        """
        Returns the derivative (df/dx) calculated at the xdatas.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.

        """
        return self.model.derivative( xdata, param )


    def setLimits( self, lowLimits=None, highLimits=None ) :
        self.model.setLimits( lowLimits=lowLimits, highLimits=highLimits )

    def getLimits( self ) :
        return self.model.getLimits()

    def nextPrior( self ) :
        yield self.model.nextPrior()

    #  ******Brackets BASENAME*******************************************************************
    def baseName( self ):
        """ Returns a string representation of the model.  """
        indent = "  "
        return "{ " + self.model._toString( indent * self.deep ) + " }"

    def basePrior( self, k ) :
        """
        Return the prior of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.
        """
        return self.model.getPrior( k )

    def hasPriors( self, isBound=True ) :
        """
        Return True when the model has priors for all its parameters.

        Parameters
        ----------
        isBound : bool
            Also check if the prior is bound.
        """
        return self.model.hasPriors( isBound=isBound )


    #  ******Brackets Parameter NAME*******************************************************************

    def baseParameterName( self, k ):
        """
        Return the name of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        return self.model.getParameterName( k )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        return self.model.getParameterUnit( k )


##### End Brackets #########################################################




import numpy as numpy
from astropy import units
import math
from . import Tools
from .Formatter import formatter as fmt
from . import Plotter
import sys
import warnings
import matplotlib.pyplot as plt

from .Problem import Problem
from .Walker import Walker
from .WalkerList import WalkerList
from .Sample import Sample
from .SampleList import SampleList
from .ErrorDistribution import ErrorDistribution

from .NestedSampler import NestedSampler

## for Order Problems import the classes
from .OrderProblem import OrderProblem
from .DistanceCostFunction import DistanceCostFunction
from .StartOrderEngine import StartOrderEngine
#from .StartNearEngine import StartNearEngine
from .MoveEngine import MoveEngine
from .LoopEngine import LoopEngine
from .ReverseEngine import ReverseEngine
from .ShuffleEngine import ShuffleEngine
from .SwitchEngine import SwitchEngine
from .NearEngine import NearEngine

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Alpha"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2019 - 2023 Do Kester

class NestedSolver( NestedSampler ):
    """
    NestedSolver is an extension of NestedSampler. It uses the
    likelihood-climbing technique to find a solution in an ordering
    problem. The negative value of the costfunction, commonly defined
    in ordering problems, is maximised. In this sense the costfunction
    is acting as the logLikelihood.

    For more information about this technique see NestedSampler.

    For the random walk of the parameters 4 so-called engines are written.
    By default only he first is switched on.

    MoveEngine    : insert a snippet of parameters at another location
    ReverseEngine : reverse the order of a snippet of parameters
    ShuffleEngine : shuffle part of the parameter list
    SwitchEngine  : switch two elements
    LoopEngine    : uncross a crossing loop
    NearEngine    : find the nearest location and go there first. 

    The last 2 engines are not random. They are mostly (?) taking steps 
    uphill. Mixing them with other engines maintain detailed balance in 
    an overall sense. 


    Attributes
    ----------
    xdata : array_like
        array of independent input values
    model : Model
        the model function to be fitted
    ydata : array_like
        array of dependent (to be fitted) data
    weights : array_like (None)
        weights pertaining to ydata
    distribution : ErrorDistribution
        to calculate the loglikelihood
    ensemble : int (100)
        number of walkers
    discard : int (1)
        number of walkers to be replaced each generation
    rng : RandomState
        random number generator
    seed : int (80409)
        seed of rng
    rate : float (1.0)
        speed of exploration
    maxsize : None or int
        maximum size of the resulting sample list (None : no limit)
    end : float (2.0)
        stopping criterion
    verbose : int
        level of blabbering

    walkers : SampleList
        ensemble of Samples that explore the likelihood space
    samples : SampleList
        Samples resulting from the exploration
    engines : list of Engine
        Engine that move the walkers around within the given constraint: logL > lowLogL
    initialEngine : Engine
        Engine that distributes the walkers over the available space
    restart : StopStart (TBW)
        write intermediate results to (optionally) start from.


    Author       Do Kester.


    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, problem, distribution=None, keep=None,
                ensemble=100, discard=1, seed=80409, rate=1.0, engines=None,
                maxsize=None, threads=False, verbose=1 ) :

        """
        Create a new class, providing inputs and model.

        Parameters
        ----------
        problem : OrderProblem
            Problem with integer parameters
        keep : None or dict of {int:float}
            None : none of the model parameters are kept fixed.
            Dictionary of indices (int) to be kept at a fixed value (float).
            Hyperparameters follow model parameters.
            The values will override those at initialization.
            They are used in this instantiation, unless overwritten at the call to sample()
        distribution : None or String or ErrorDistribution
            None   : DistanceCostFunction is chosen.

            "distance" : `DistanceCostFunction`      no hyperpar

            errdis : A class inheriting from ErrorDistribution
                     which implements logLikelihood

            When the hyperpar(s) are not to be kept fixed, they need `Prior` and maybe limits.
        ensemble : int (100)
            number of walkers
        discard : int (1)
            number of walkers to be replaced each generation
        seed : int (80409)
            seed of rng
        rate : float (1.0)
            speed of exploration
        engines : None or (list of) string or (list of) Engine
            to randomly move the walkers around, within the likelihood bound.

            "move"    : insert a snippet of parameters at another location
            "reverse" : reverse the order of a snippet of parameters
            "shuffle" : shuffle part of the parameter list
            "switch"  : switch two elements
            "loop"    : find two paths that cross, then uncross them
            "near"    : find the nearest location and go there first. 

            None    : take default [all of above].

            engine  : a class inheriting from Engine. At least implementing
                      execute( walker, lowLhood )
        maxsize : None or int
            maximum size of the resulting sample list (None : no limit)
        threads : bool (False)
            Use Threads to distribute the diffusion of discarded samples over the available cores.
        verbose : int (1)
            0 : silent
            1 : basic information
            2 : more about every 100th iteration
            3 : more about every iteration

        """
        if distribution is None :
            self.setErrorDistribution( "distance" )


        super().__init__( problem=problem, distribution=distribution, keep=keep,
                ensemble=ensemble, discard=discard, seed=seed, rate=rate,
                engines=engines, maxsize=maxsize, threads=threads, verbose=verbose )

        self.setEngines( engines )


    #  *******SAMPLE************************************************************
    def solve( self, keep=None, plot=False ):
        """
        Solve an order problem.

        Return the last sample, representing the best solution.

        The more sammples (with solutions) can be found in the sample list.

        Parameters
        ----------
        keep : None or dict of {int:float}
            Dictionary of indices (int) to be kept at a fixed value (float)
            Hyperparameters follow model parameters
            The values will override those at initialization.
            They are only used in this call of fit.
        plot : bool
            Show a plot of the results

        """
        self.sample( keep=keep, plot=plot )

        return self.samples[-1]


    #  *********DISTRIBUTIONS***************************************************
    def setErrorDistribution( self, name=None, scale=1.0, power=2.0 ):
        """
        Set the error distribution for calculating the likelihood.

        Parameters
        ----------
        name : string
            name of distribution
        scale : float
            fixed scale of distribution
        power : float
            fixed power of distribution

        """
        if name is None :
            name = self.problem.myDistribution()
        elif isinstance( name, ErrorDistribution ) :
            self.distribution = name

            return

        if not isinstance( name, str ) :
            raise ValueError( "Cannot interpret ", name, " as string or ErrorDistribution" )

        name = str.lower( name )
#        print( name )
        if name == "distance" :
            self.distribution = DistanceCostFunction( )
        else :
            raise ValueError( "Unknown error distribution %s" % name )

    def setEngines( self, engines=None, enginedict=None ) :
        """
        initialize the engines.

        Parameters
        ----------
        engines : list of string
            list of engine names
        enginedict : dictionary of { str : Engine }
            connecting names to Engines

        """
        if enginedict is None :
            enginedict = {
               "move"    : MoveEngine,
               "reverse" : ReverseEngine,
               "loop"    : LoopEngine,
               "near"    : NearEngine,
               "shuffle" : ShuffleEngine,
               "switch"  : SwitchEngine }

        super().setEngines( engines=engines, enginedict=enginedict )

    #  *********INITIALIZATION***************************************************

    def initWalkers( self, ensemble, allpars, fitIndex, startdict=None ):
        """
        Initialize the walkers at random values of parameters and scale

        Parameters
        ----------
        ensemble : int
            length od the walkers list
        allpars : array_like
            array of parameters
        fitIndex : array_like
            indices of allpars to be fitted
        startdict : dictionary of { str : Engine }
            connecting a name to a StartEngine
        """
        if startdict is None :
            startdict = { "startorder" : StartOrderEngine }

#        print( "initW  ", allpars, fitIndex )

        allpars = numpy.arange( len( allpars ), dtype=int )

        super().initWalkers( ensemble, allpars, fitIndex, startdict=startdict )





